__all__ = ('URL', )

from itertools import islice
from warnings import warn

from ...utils import RichAttributeErrorBaseType, cached_property

from ..quoting import quote

from .constants import DEFAULT_PORTS, HOST_RP, URL_RP
from .url_fragment import URLFragment
from .url_host_info import URLHostInfo
from .url_path import normalize_path_parts, URLPath
from .url_query import URLQuery, normalize_query
from .url_user_info import URLUserInfo


def _create_host(iterator, encoded):
    """
    Creates an url's host info from the given iterator.
    
    Parameters
    ----------
    iterator : `iterator<None | str>`
        Iterator to go over required amount of elements of it.
    
    encoded : `bool`
        Whether the given `value` is already encoded.
    
    Returns
    -------
    host : `None | URLHostInfo`
    """
    host_local_ip_v4 = next(iterator)
    host_local_name = next(iterator)
    host_local_unambiguous = next(iterator)
    host_external_ip_v4 = next(iterator)
    host_external_ip_v6 = next(iterator)
    host_external_sub_domain_name = next(iterator)
    host_external_main_domain_name = next(iterator)
    host_external_top_level_domain_name = next(iterator)
    host_external_unambiguous = next(iterator)
    
    if (host_local_ip_v4 is not None):
        if encoded:
            host = URLHostInfo.create_from_local_ip_v4_encoded(host_local_ip_v4)
        else:
            host = URLHostInfo.create_from_local_ip_v4_decoded(host_local_ip_v4)
    
    elif (host_local_name is not None):
        unambiguous = (host_local_unambiguous is not None)
        if encoded:
            host = URLHostInfo.create_from_local_name_encoded(host_local_name, unambiguous)
        else:
            host = URLHostInfo.create_from_local_name_decoded(host_local_name, unambiguous)
    
    elif (host_external_ip_v4 is not None):
        if encoded:
            host = URLHostInfo.create_from_external_ip_v4_encoded(host_external_ip_v4)
        else:
            host = URLHostInfo.create_from_external_ip_v4_decoded(host_external_ip_v4)
    
    elif (host_external_ip_v6 is not None):
        if encoded:
            host = URLHostInfo.create_from_external_ip_v6_encoded(host_external_ip_v6)
        else:
            host = URLHostInfo.create_from_external_ip_v6_decoded(host_external_ip_v6)
    
    elif (
        (host_external_sub_domain_name is not None) or
        (host_external_main_domain_name is not None) or
        (host_external_top_level_domain_name is not None)
    ):
        unambiguous = (host_external_unambiguous is not None)
        if encoded:
            host = URLHostInfo.create_from_external_name_encoded(
                host_external_sub_domain_name,
                host_external_main_domain_name,
                host_external_top_level_domain_name,
                unambiguous,
            )
        else:
            host = URLHostInfo.create_from_external_name_decoded(
                host_external_sub_domain_name,
                host_external_main_domain_name,
                host_external_top_level_domain_name,
                unambiguous,
            )
    
    else:
        host = None
    
    return host


class URL(RichAttributeErrorBaseType):
    """
    Represents an url (Uniform Resource Locator).
    
    Attributes
    ----------
    _cache : `None | dict<str, object>`
        Cache.
    
    _fragment : `None | URLFragment`
        The url's fragment component.
    
    _host : `None | URLHostInfo`
        The url's host component.
    
    _path : `None | URLPath`
        The url's path component.
    
    _port : `None | int`
        The url's port.
    
    _query : `None | URLQuery`
        The url's query component.
    
    _scheme : `None | str`
        The url's scheme.
    
    _user : `None | URLUserInfo``
        The url's user component.
    """
    __slots__ = ('_cache', '_fragment', '_host', '_path', '_port', '_query', '_scheme', '_user')
    
    def __new__(cls, value = '', encoded = False):
        """
        Creates a new url from the given value.
        
        Parameters
        ----------
        value : `str | instance<cls>` = `''`, Optional
            The value to create instance from.
        
        encoded : `bool` = `False`, Optional
            Whether the given `value` is already encoded.
        
        Raises
        -------
        ValueError
            - If a parameter's value is incorrect.
        TypeError
            - If a parameter's type is incorrect.
        """
        if isinstance(value, cls):
            return value
        
        if isinstance(value, str):
            if not value:
                cls._create_empty()
            
            return cls._create_from_string(value, encoded)
            
        raise TypeError(
            f'`value` can be `{cls.__name__}`, `str`. Got '
            f'{type(value).__name__}; {value!r}.'
        )
    
    
    @classmethod
    def _create_empty(cls):
        """
        Creates an empty url.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self._cache = None
        self._fragment = None
        self._host = None
        self._path = None
        self._port = None
        self._query = None
        self._scheme = None
        self._user = None
        return self
    
    
    @classmethod
    def _create_from_string(cls, value, encoded):
        """
        Creates an url from the given string.
        
        Parameters
        ----------
        value : `str`
            the value to create url from.
        
        encoded : `bool`
            Whether the given `value` is already encoded.
        
        Returns
        -------
        self : `instance<cls>`
        """
        match = URL_RP.fullmatch(value)
        if match is None:
            raise ValueError(
                f'The given value is not an url: {value!r}.'
            )
        
        iterator = iter(match.groups())
        
        # scheme
        scheme = next(iterator)
        # nothing to do, we expected it to be already in good format
        
        # user
        user_name = next(iterator)
        user_password = next(iterator)
        if (user_name is not None) or (user_password is not None):
            if encoded:
                user = URLUserInfo.create_from_encoded(user_name, user_password)
            else:
                user = URLUserInfo.create_from_decoded(user_name, user_password)
        
        else:
            user = None
        
        # host
        host = _create_host(iterator, encoded)
        
        # port
        port = next(iterator)
        if (port is not None):
            port = int(port)
        
        # path
        path = next(iterator)
        if (path is not None):
            if encoded:
                path = URLPath.create_from_encoded(path)
            else:
                path = URLPath.create_from_decoded(path)
        
        # query
        query = next(iterator)
        if (query is not None):
            if encoded:
                query = URLQuery.create_from_encoded(query)
            else:
                query = URLQuery.create_from_decoded(query)
        
        # fragment
        fragment = next(iterator)
        if (fragment is not None):
            if encoded:
                fragment = URLFragment.create_from_encoded(fragment)
            else:
                fragment = URLFragment.create_from_decoded(fragment)
        
        # Construct
        self = object.__new__(cls)
        self._cache = None
        self._fragment = fragment
        self._host = host
        self._path = path
        self._port = port
        self._query = query
        self._scheme = scheme
        self._user = user
        return self
    
    
    @cached_property
    def value_decoded(self):
        """
        Returns the decoded value of the url.
        
        Returns
        -------
        value : `str`
        """
        return self._build_value(False)
    
    
    @cached_property
    def value_encoded(self):
        """
        Returns the encoded value of the url.
        
        Returns
        -------
        value : `str`
        """
        return self._build_value(True)
    
    
    def _build_value(self, encoded):
        """
        Builds the string value of the url.
        
        Parameters
        ----------
        encoded : `bool`
            Whether should build an encoded value.
        """
        built = []
        
        # scheme
        scheme = self._scheme
        if (scheme is not None):
            built.append(scheme)
            built.append(':')
        
        # authority (can be missing if relative)
        user = self._user
        port = self._port
        if (user is not None) or (self._host is not None) or (port is not None):
            if built:
                built.append('//')
            
            # user
            if (user is not None):
                if encoded:
                    user_name = user.name_encoded
                    user_password = user.password_encoded
                else:
                    user_name = user.name_decoded
                    user_password = user.password_decoded
                
                if (user_name is not None):
                    built.append(user_name)
                
                built.append(':')
                
                if (user_password is not None):
                    built.append(user_password)
                
                built.append('@')
            
            # host
            built = self._build_host_into(built, encoded)
            
            # port
            if (port is not None):
                built.append(':')
                built.append(str(port))
        
        
        # path
        path = self._path
        if (path is None):
            path_value = None
        else:
            if encoded:
                path_value = path.encoded
            else:
                path_value = path.decoded
        
        if path_value is None:
            path_value = '/'
        
        built.append(path_value)
        
        # query
        query = self._query
        if (query is not None):
            if encoded:
                query_value = query.encoded
            else:
                query_value = query.decoded
            
            if (query_value is not None):
                built.append('?')
                built.append(query_value)
        
        # fragment
        fragment = self._fragment
        if (fragment is not None):
            if encoded:
                fragment_value = fragment.encoded
            else:
                fragment_value = fragment.decoded
            
            if (fragment_value is not None):
                built.append('#')
                built.append(fragment_value)
        
        # Remove `/` if we have only that.
        if len(built) == 1 and built[0] == '/':
            return ''
        
        return ''.join(built)
    
    
    def _build_host_into(self, into, encoded):
        """
        Builds the host section of the url into the given container.
        
        Parameters
        ----------
        into : `list<str>`
            Container to build into.
        
        encoded : `bool`
            Whether should build encoded authority.
        
        Returns
        -------
        into : `list<str>`
        """
        host = self._host
        # host (should never be None if we have this case)
        if (host is not None):
            if encoded:
                host_main_domain = host.main_domain_encoded
                host_sub_domain = host.sub_domain_encoded
                host_top_level_domain = host.top_level_domain_encoded
            else:
                host_main_domain = host.main_domain_decoded
                host_sub_domain = host.sub_domain_decoded
                host_top_level_domain = host.top_level_domain_decoded
            
            if (host_sub_domain is not None):
                into.append(host_sub_domain)
                into.append('.')
            
            # (should never be None)
            if (host_main_domain is not None):
                ip_v6 = host.is_ip_v6()
                if ip_v6:
                    into.append('[')
                
                into.append(host_main_domain)
                
                if ip_v6:
                    into.append(']')
            
            if (host_top_level_domain is not None):
                into.append('.')
                into.append(host_top_level_domain)
            
            if host.is_unambiguous():
                into.append('.')
        
        
        return into
    
    
    def __str__(self):
        """Returns str(self)."""
        return self.value_decoded
    
    
    def __repr__(self):
        """Returns repr(self)."""
        return f'{type(self).__name__}({self.value_decoded!r})'
    
    
    def __hash__(self):
        """Returns hash(self)."""
        return hash(self.value_encoded)
    
    
    def __bool__(self):
        """Returns bool(self)."""
        # fragment
        if (self._fragment is not None):
            return True
        
        # host
        if (self._host is not None):
            return True
        
        # path
        if (self._path is not None):
            return True
        
        # port
        if (self._port is not None):
            return True
        
        # query
        if (self._query is not None):
            return True
        
        # scheme
        if (self._scheme is not None):
            return True
        
        # user
        if (self._user is not None):
            return True
        
        return False
        
    
    def __gt__(self, other):
        """Returns (self > other)."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self.value_encoded > other.value_encoded
    
    
    def __ge__(self, other):
        """Returns (self >= other)."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self.value_encoded >= other.value_encoded
        
        
    def __eq__(self, other):
        """Returns (self == other)."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self.value_encoded == other.value_encoded
    
    
    def __ne__(self, other):
        """Returns (self != other)."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self.value_encoded != other.value_encoded
    
    
    def __le__(self, other):
        """Returns (self <= other)."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self.value_encoded <= other.value_encoded
    
    
    def __lt__(self, other):
        """Returns (self < other)."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self.value_encoded < other.value_encoded
    
    
    def __truediv__(self, name):
        """Returns self / other"""
        # Join *self.path with name
        path = self._path
        if (path is None):
            split = None
        else:
            split = path.parsed
        
        if split is None:
            split = [name]
        else:
            split = [*split, name]
        
        split = normalize_path_parts(split)
        if split is None:
            path = None
        else:
            path = URLPath.create_from_parsed(tuple(split))
        
        # Construct
        new = object.__new__(type(self))
        new._cache = None
        new._fragment = self._fragment
        new._host = self._host
        new._path = path
        new._port = self._port
        new._query = self._query
        new._scheme = self._scheme
        new._user = self._user
        return new
    
    
    def is_absolute(self):
        """
        Returns whether the url is absolute (having scheme or starting with //).
        
        Returns
        -------
        is_absolute : `bool`
        """
        return (self._host is not None)
    
    
    def is_default_port(self):
        """
        Returns whether the url's port is default, like: 'https://orindance.party/' or 'https://orindance.party:80/'.
        
        Returns
        -------
        is_default_port : `bool`
        """
        scheme = self._scheme
        if scheme is None:
            return False
        
        scheme_default_port = DEFAULT_PORTS.get(scheme, None)
        if scheme_default_port is None:
            return False
        
        port = self._port
        
        if port is None:
            port = DEFAULT_PORTS.get(scheme, None)
            if port is None:
                return False
        
        return scheme_default_port == port
    
    
    def is_host_ip(self):
        """
        Returns whether the host is an ip address.
        
        Returns
        -------
        is_ip : `bool`
        """
        host = self._host
        if host is None:
            return False
        
        return host.is_ip()
    
    
    def is_host_ip_v4(self):
        """
        Returns whether the host is an ipv4 address.
        
        Returns
        -------
        is_ip_v4 : `bool`
        """
        host = self._host
        if host is None:
            return False
        
        return host.is_ip_v4()
    
    
    def is_host_ip_v6(self):
        """
        Returns whether the host is an ipv6 address.
        
        Returns
        -------
        is_ip_v6 : `bool`
        """
        host = self._host
        if host is None:
            return False
        
        return host.is_ip_v6()
    
    
    def origin(self):
        """
        Returns an url with scheme, host and port parts only, user, password, path, query and fragment are removed.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        ValueError
            - If the url is not absolute.
        """
        if not self.is_absolute():
            raise ValueError(
                f'Url should be absolute; self = {self!r}.'
            )
        
        # Construct
        new = object.__new__(type(self))
        new._cache = None
        new._fragment = None
        new._host = self._host
        new._path = None
        new._port = self._port
        new._query = None
        new._scheme = self._scheme
        new._user = None
        return new

    
    def relative(self):
        """
        Returns a relative part of the url. Scheme, user, password, host and port are removed.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        ValueError
            If url is not absolute.
        """
        if not self.is_absolute():
            raise ValueError(
                f'Url should be absolute; self = {self!r}.'
            )
        
        # Construct
        new = object.__new__(type(self))
        new._cache = None
        new._fragment = self._fragment
        new._host = None
        new._path = self._path
        new._port = None
        new._query = self._query
        new._scheme = None
        new._user = None
        return new
    
    
    @property
    def scheme(self):
        """
        Returns the scheme for absolute url-s. Returns `None` for relatives or if the url has no scheme.
        
        Returns
        -------
        scheme : `None | str`
        """
        return self._scheme

    
    @property
    def raw_user(self):
        """
        Returns the encoded user part of the url. `None` is returned if the user part is missing.
        
        Returns
        -------
        raw_user : `None | str`
        """
        user = self._user
        if (user is not None):
            return user.name_encoded
    
    
    @property
    def user(self):
        """
        Returns the decoded user part of the url. `None` is returned if the user part is missing.
        
        Returns
        -------
        user : `None | str`
        """
        user = self._user
        if (user is not None):
            return user.name_decoded
    
    
    @property
    def raw_password(self):
        """
        Returns the encoded password part of the url. Returns `None` if password is missing.
        
        Returns
        -------
        raw_password : `None | str`
        """
        user = self._user
        if (user is not None):
            return user.password_encoded
    
    
    @property
    def password(self):
        """
        Returns the decoded password part of the url. Returns `None` if password is missing.
        
        Returns
        -------
        password : `None | str`
        """
        user = self._user
        if (user is not None):
            return user.password_decoded

    
    @cached_property
    def raw_host(self):
        """
        Returns the encoded host part of the url. Returns `None` if the host part is missing or if the url is relative.
        
        Returns
        -------
        raw_host : `None | str`
        """
        built = self._build_host_into([], True)
        if built:
            return ''.join(built)
    
    
    @cached_property
    def host(self):
        """
        Returns the decoded host part of the url. Returns `None` if the host part is missing or if the url is relative.
        
        Returns
        -------
        host : `None`, `str`
        """
        built = self._build_host_into([], False)
        if built:
            return ''.join(built)
    
    
    @property
    def raw_subdomain(self):
        """
        Deprecated and will be removed in 2025 November. Use `.raw_sub_domain` instead.
        """
        warn(
            (
                f'`{type(self).__name__}.raw_subdomain` is deprecated and will be removed in 2025 November. '
                f'Please use `.raw_sub_domain` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        return self.raw_sub_domain
    
    
    @property
    def subdomain(self):
        """
        Deprecated and will be removed in 2025 November. Use `.sub_domain` instead.
        """
        warn(
            (
                f'`{type(self).__name__}.subdomain` is deprecated and will be removed in 2025 November. '
                f'Please use `.sub_domain` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        return self.sub_domain
    
    
    @property
    def raw_sub_domain(self):
        """
        Returns the encoded sub domain part of the url. Returns `None` if the sub domain part is missing or if
        the url is relative.
        
        Returns
        -------
        raw_sub_domain : `None | str`
        """
        host = self._host
        if (host is not None):
            return host.sub_domain_encoded
    
    
    @property
    def sub_domain(self):
        """
        Returns the decoded sub domain part of the url. Returns `None` if the sub domain part is missing or if
        the url is relative.
        
        Returns
        -------
        sub_domain : `None`, `str`
        """
        host = self._host
        if (host is not None):
            return host.sub_domain_decoded

    
    @property
    def port(self):
        """
        Returns the port part of url. Returns `None` if the url is relative, if the url not contains port part and if
        the port can's be detected from the url's scheme.
        
        Returns
        -------
        port : `None | int`
        """
        port = self._port
        if (port is not None):
            return port
        
        scheme = self._scheme
        if (scheme is not None):
            return DEFAULT_PORTS.get(scheme, None)
    
    
    @property
    def raw_path(self):
        """
        Returns the encoded path part of the url. Returns `None` if the url has no path.
        
        Returns
        -------
        raw_path : `str`
        """
        path = self._path
        if (path is None):
            path_value = None
        else:
            path_value = path.encoded
        
        if path_value is None:
            path_value = '/'
        
        return path_value
    
    
    @property
    def path(self):
        """
        Returns the decoded path part of the url. Returns `None` if the url has no path.
        
        Returns
        -------
        path : `str`
        """
        path = self._path
        if (path is None):
            path_value = None
        else:
            path_value = path.decoded
        
        if path_value is None:
            path_value = '/'
        
        return path_value
    
    
    @property
    def query(self):
        """
        Returns a multi value dictionary representing parsed query parameters in decoded representation.
        
        `None` if returned if the url has no query.
        
        Returns
        -------
        query : `None | MultiValueDictionary<str, str>`
        """
        query = self._query
        if (query is not None):
            return query.parsed
    
    
    @property
    def raw_query_string(self):
        """
        Returns the encoded query string part of url. Returns an `None` if the query part is missing.
        
        Returns
        -------
        raw_query_string : `None | str`
        """
        query = self._query
        if (query is not None):
            return query.encoded
    
    
    @property
    def query_string(self):
        """
        Returns the decoded query string part of url. Returns `None` if the query part is missing.
        
        Returns
        -------
        query_string : `str`
        """
        query = self._query
        if (query is not None):
            return query.decoded
    
    
    @property
    def raw_fragment(self):
        """
        Returns the encoded fragment part of the url. Returns `None` if the fragment part is missing.
        
        Returns
        -------
        raw_fragment : `None | str`
        """
        fragment = self._fragment
        if (fragment is not None):
            return fragment.encoded
    
    
    @property
    def fragment(self):
        """
        Returns the decoded fragment part of the url. Returns `None` if the fragment part is missing.
        
        Returns
        -------
        raw_fragment : `None | str`
        """
        fragment = self._fragment
        if (fragment is not None):
            return fragment.decoded
    
    
    @cached_property
    def raw_parts(self):
        """
        Deprecated and will be removed in 2025 November.
        """
        warn(
            f'`{type(self).__name__}.raw_parts` is deprecated and will be removed in 2025 November.',
            FutureWarning,
            stacklevel = 2,
        )
        path = self._path
        if (path is not None):
            return (*(quote(part, safe = '@:', protected = '/') for part in path.parsed),)
    
    
    @property
    def parts(self):
        """
        Returns a `tuple` containing decoded path parts. If the url has no path returns `None`.

        Returns
        -------
        parts : `None | tuple<str>`
        """
        path = self._path
        if (path is not None):
            return path.parsed
    
    
    @cached_property
    def parent(self):
        """
        Returns a new url with last part of path removed without query and fragment.
        
        Returns
        -------
        parent : `instance<type<self>>`
        """
        path = self._path
        if (path is not None):
            path_parts = path.parsed
            if (path_parts is None):
                path = None
            
            elif len(path_parts) == 1:
                path = None
            
            else:
                path = URLPath.create_from_parsed(path_parts[:-1])
        
        # Construct
        new = object.__new__(type(self))
        new._cache = None
        new._fragment = self._fragment
        new._host = self._host
        new._path = path
        new._port = self._port
        new._query = None
        new._scheme = self._scheme
        new._user = self._user
        return new
    
    
    @cached_property
    def raw_name(self):
        """
        Deprecated and will be removed in 2025 November.
        """
        warn(
            f'`{type(self).__name__}.raw_parts` is deprecated and will be removed in 2025 November.',
            FutureWarning,
            stacklevel = 2,
        )
        path = self._path
        if (path is not None):
            path_parts = path.parsed
            if (path_parts is not None):
                return quote(path_parts[-1], safe = '@:', protected = '/')
    
    
    @cached_property
    def name(self):
        """
        Returns the last part of ``.parts``. If there are no parts, returns `None`.
        
        Returns
        -------
        raw_name : `str`
        """
        path = self._path
        if (path is not None):
            path_parts = path.parsed
            if (path_parts is not None):
                return path_parts[-1]
        
        return ''
    
    
    def with_scheme(self, scheme):
        """
        Returns a new url with `scheme` replaced.
        
        Parameters
        ----------
        scheme : `None | str`
            Scheme part for the new url.

        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            If `scheme` was not given as `None`, `str`.
        ValueError
            If the source url is relative; Scheme replacement is not allowed for relative url-s.
        
        Notes
        -----
        The returned url's `query` and `fragment` will be same as the source one's.
        """
        if (scheme is None):
            pass
        
        elif isinstance(scheme, str):
            if not scheme:
                scheme = None
        
        else:
            raise TypeError(
                f'`scheme` can be `None`, `str`, got {type(scheme).__name__}; {scheme!r}.'
            )
        
        if not self.is_absolute():
            raise ValueError(
                f'`scheme` replacement is not allowed for relative url-s; self = {self!r}; scheme = {scheme!r}.'
            )
        
        # Construct
        new = object.__new__(type(self))
        new._cache = None
        new._fragment = self._fragment
        new._host = self._host
        new._path = self._path
        new._port = self._port
        new._query = self._query
        new._scheme = scheme
        new._user = self._user
        return new
    
    
    def with_user(self, user_name):
        """
        Returns a new url with `user` replaced.
        
        Parameters
        ----------
        user_name : `None | str`
            New user name part for the new url.

        Returns
        -------
        new : `instance<type<cls>>`
        
        Raises
        ------
        TypeError
            If `user` was not given neither as `None` nor `str`.
        ValueError
            If the source url is relative; User replacement is not allowed for relative url-s.
        
        Notes
        -----
        The returned url's `query` and `fragment` will be same as the source one's.
        """
        if user_name is None:
            user_password = self.password
        
        elif isinstance(user_name, str):
            if not user_name:
                user_name = None
            
            user_password = self.password
            
        else:
            raise TypeError(
                f'`user_name` can be `None`, `str`, got {type(user_name).__name__}; {user_name!r}.'
            )
        
        if not self.is_absolute():
            raise ValueError(
                f'`user` replacement is not allowed for relative url-s; self = {self!r}; user_name = {user_name!r}.'
            )
        
        if (user_name is None) and (user_password is None):
            user = None
        else:
            user = URLUserInfo.create_from_decoded(user_name, user_password)
        
        # Construct
        new = object.__new__(type(self))
        new._cache = None
        new._fragment = self._fragment
        new._host = self._host
        new._path = self._path
        new._port = self._port
        new._query = self._query
        new._scheme = self._scheme
        new._user = user
        return new
    
    
    def with_password(self, user_password):
        """
        Returns a new url with `password` replaced.
        
        Give `password` as `None` to clear it from the source url.
        
        Parameters
        ----------
        user_password : `None | str`
            New user password part for the new url.

        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            If `password` was not given neither as `None` nor `str`.
        ValueError
            If the source url is relative; Password replacement is not allowed for relative url-s.
        
        Notes
        -----
        The returned url's `query` and `fragment` will be same as the source one's.
        """
        if (user_password is None):
            user_name = self.user
        
        elif isinstance(user_password, str):
            user_name = self.user
            
            if not user_password:
                user_password = None
        
        else:
            raise TypeError(
                f'`user_password` can be `None`, `str`, got {type(user_password).__name__}; {user_password!r}.'
            )
        
        if not self.is_absolute():
            raise ValueError(
                f'Password replacement is not allowed for relative url-s; '
                f'self = {self!r}; user_password = {user_password}.'
            )
        
        if (user_name is None) and (user_password is None):
            user = None
        else:
            user = URLUserInfo.create_from_decoded(user_name, user_password)
        
        # Construct
        new = object.__new__(type(self))
        new._cache = None
        new._fragment = self._fragment
        new._host = self._host
        new._path = self._path
        new._port = self._port
        new._query = self._query
        new._scheme = self._scheme
        new._user = user
        return new
    

    def with_host(self, host):
        """
        Returns a new url with `host` replaced.
        
        Parameters
        ----------
        host : `str`
            Host part for the new url.

        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            If `host` was not given as `str`.
        ValueError
            - If the source url is relative; Host replacement is not allowed for relative url-s.
            - If `host` was given as empty string; Removing `host` is not allowed.
            - If `host` could not be matched to any patterns.
        
        Notes
        -----
        The returned url's `query` and `fragment` will be same as the source one's.
        """
        if not isinstance(host, str):
            raise TypeError(
                f'`host` can be `str`, got {type(host).__name__}; {host!r}.'
            )
        
        if not host:
            raise ValueError(
                f'`host` was given as empty string, but removing host is not allowed; self = {self!r}.'
            )
        
        if not self.is_absolute():
            raise ValueError(
                f'`host` replacement is not allowed for relative url-s; self = {self!r}; host = {host!r}.'
            )
        
        match = HOST_RP.fullmatch(host)
        if match is None:
            raise ValueError(
                f'`host` could not be matched to any known pattern: host = {host!r}'
            )
        
        iterator = iter(match.groups())
        host = _create_host(iterator, False)
        
        # Construct
        new = object.__new__(type(self))
        new._cache = None
        new._fragment = self._fragment
        new._host = host
        new._path = self._path
        new._port = self._port
        new._query = self._query
        new._scheme = self._scheme
        new._user = self._user
        return new
    
    
    def with_port(self, port):
        """
        Returns a new url with `port` replaced. Give `port` as None` to clear it to default.
        
        Parameters
        ----------
        port : `None | int`
            Port part of the new url.

        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If `port` was not given neither as `None` nor `int`.
        ValueError
            - If the source url is relative; Port replacement is not allowed for relative url-s.
            - If `port` is negative.
        
        Notes
        -----
        The returned url's `query` and `fragment` will be same as the source one's.
        """
        if (port is None):
            pass
        
        elif isinstance(port, int):
            if port < 0:
                raise ValueError(
                    f'`port` cannot be negative, got {port!r}.'
                )
        
        else:
            raise TypeError(
                f'`port` can be `None`, `int`, got {type(port).__name__}; {port!r}.'
            )
        
        if not self.is_absolute():
            raise ValueError(
                f'`port` replacement is not allowed for relative url-s; self = {self!r}; port = {port}.'
            )
        
        # Construct
        new = object.__new__(type(self))
        new._cache = None
        new._fragment = self._fragment
        new._host = self._host
        new._path = self._path
        new._port = port
        new._query = self._query
        new._scheme = self._scheme
        new._user = self._user
        return new
    
    
    def with_query(self, query):
        """
        Returns a new url with query part replaced. By giving `None` you can clear the actual query.
        
        Parameters
        ----------
        query : `None | str | dict<str, str | int | bool | None | float | DateTime | (list | set)<...>> | (list | set)<(str, ...)>`
            The query to use.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If `query` was given as an invalid type.
            - If `query` was given as `set`, `list`, but `1` of it's elements cannot be unpacked correctly.
            - If a query key was not given as `str`.
            - If a query value was not given as any of the expected types.
        ValueError
            - If a query value was given as `float`, but as `inf`.
            - If a query value was given as `float`, but as `nan`.
        
        Notes
        -----
        The returned url's `fragment` will be same as the source one's.
        """
        query = normalize_query(query)
        if (query is not None):
            query = URLQuery.create_from_parsed(query)
        
        # Construct
        new = object.__new__(type(self))
        new._cache = None
        new._fragment = self._fragment
        new._host = self._host
        new._path = self._path
        new._port = self._port
        new._query = query
        new._scheme = self._scheme
        new._user = self._user
        return new

    
    def with_fragment(self, fragment):
        """
        Returns a new url with `fragment` replaced. Give `fragment` as None` to clear it.
        
        Parameters
        ----------
        fragment : `None | str`
            Fragment part of the new url.

        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            If `fragment` was not given neither as `None` nor `str`.
        
        Notes
        -----
        The returned url's `query` will be same as the source one's.
        """
        if fragment is None:
            pass
        
        elif isinstance(fragment, str):
            if not fragment:
                fragment = None
        
        else:
            raise TypeError(
                f'`fragment` can be `None`, `str`, got {type(fragment).__name__}; {fragment!r}.'
            )
        
        if (fragment is not None):
            fragment = URLFragment.create_from_decoded(fragment)
        
        # Construct
        new = object.__new__(type(self))
        new._cache = None
        new._fragment = fragment
        new._host = self._host
        new._path = self._path
        new._port = self._port
        new._query = self._query
        new._scheme = self._scheme
        new._user = self._user
        return new
    
    
    def with_name(self, name):
        """
        Returns a new url with `name` (last part of path) replaced.
        
        Parameters
        ----------
        name : `str`
            Name part of the new url.

        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            If `name` was not given as `str`.
        
        Notes
        -----
        The returned url's `fragment` and `query` will be REMOVED.
        """
        if not isinstance(name, str):
            raise TypeError(
                f'`name` can be `str`, got {type(name).__name__}; {name!r}.'
            )
        
        path = self._path
        if path is None:
            path_parts = None
        else:
            path_parts = path.parsed
        
        if path_parts is None:
            path_parts_length = 0
        else:
            path_parts_length = len(path_parts)
        
        if path_parts_length <= 1:
            path_parts = [name]
        
        else:
            path_parts = [*islice(path_parts, 0, path_parts_length - 1), name]
        
        path_parts = normalize_path_parts(path_parts)
        if (path_parts is not None):
            path_parts = tuple(path_parts)
        
        if path_parts is None:
            path = None
        else:
            path = URLPath.create_from_parsed(path_parts)
        
        # Construct
        new = object.__new__(type(self))
        new._cache = None
        new._fragment = self._fragment
        new._host = self._host
        new._path = path
        new._port = self._port
        new._query = self._query
        new._scheme = self._scheme
        new._user = self._user
        return new
    
    
    def join(self, other):
        """
        Joins two urls.
        
        Construct a full ('absolute') url by combining a 'base url' (self) with another one (other).
        
        Informally, this uses components of the base url, in particular the addressing scheme, the network location and
        (part of) the path, to provide missing components in the relative url.
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other url to join to self.
        
        Raises
        ------
        TypeError
            If `other` was not given as the same type.
        """
        if not isinstance(other, type(self)):
            raise TypeError(
                f'`url` can be `{type(self).__name__}`, got {type(other).__name__}; {other!r}.'
            )
        
        self_path = self._path
        if self_path is None:
            self_path_parts = None
        else:
            self_path_parts = self_path.parsed
        
        other_path = other._path
        if other_path is None:
            other_path_parts = None
        else:
            other_path_parts = other_path.parsed
        
        if self_path_parts is None:
            path_parts = other_path_parts
        
        elif other_path_parts is None:
            path_parts = self_path_parts
        
        else:
            path_parts = [*self_path_parts, *other_path_parts]
            path_parts = normalize_path_parts(path_parts)
            if (path_parts is not None):
                path_parts = tuple(path_parts)
        
        if path_parts is None:
            path = None
        else:
            path = URLPath.create_from_parsed(path_parts)
        
        # Construct
        new = object.__new__(type(self))
        new._cache = None
        new._fragment = other._fragment
        new._host = self._host
        new._path = path
        new._port = self._port
        new._query = other._query
        new._scheme = self._scheme
        new._user = self._user
        return new
    
    
    def extend_query(self, query):
        """
        Returns a new url with it's query parameters extended.
        
        Parameters
        ----------
        query : `None | str | dict<str, str | int | bool | None | float | DateTime | (list | set)<...>> | (list | set)<(str, ...)>`
            The query to use.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If `query` was given as an invalid type.
            - If `query` was given as `set`, `list`, but `1` of it's elements cannot be unpacked correctly.
            - If a query key was not given as `str`.
            - If a query value was not given as any of the expected types.
        ValueError
            - If a query value was given as `float`, but as `inf`.
            - If a query value was given as `float`, but as `nan`.
        
        Notes
        -----
        The returned url's `fragment` will be same as the source one's.
        """
        other_query = normalize_query(query)
        
        query = self._query
        if query is None:
            self_query = None
        else:
            self_query = query.parsed
        
        if other_query is None:
            if self_query is None:
                query = None
            else:
                query = self_query.copy()
        
        else:
            if self_query is None:
                query = other_query
            else:
                query = self_query.copy()
                query.extend(other_query)
        
        if (query is not None):
            query = URLQuery.create_from_parsed(query)
        
        # Construct
        new = object.__new__(type(self))
        new._cache = None
        new._fragment = self._fragment
        new._host = self._host
        new._path = self._path
        new._port = self._port
        new._query = query
        new._scheme = self._scheme
        new._user = self._user
        return new



__all__ = ()

from socket import AI_NUMERICHOST as ADDRESS_INFO_NUMERIC_HOST

from ..utils import RichAttributeErrorBaseType


class HostInfo(RichAttributeErrorBaseType):
    """
    Resolved information about a host.
    
    Attributes
    ----------
    family : `AddressFamily`
        Address family.
    
    flags : `int`
        Bit-mask for `get_address_info`.
    
    host : `str`
        The host's ip address.
    
    host_name : `str`
        The hosts' name.
    
    port : `int`
        Port to connect to the host.
    
    protocol : `int`
        Protocol type.
    """
    __slots__ = ('family', 'flags', 'host', 'host_name', 'port', 'protocol')
    
    def __new__(cls, family, flags, host, host_name, port, protocol):
        """
        Creates a new host info.
        
        Parameters
        ----------
        family : `AddressFamily`
            Address family.
        
        flags : `int`
            Bit-mask for `get_address_info`.
        
        host : `str`
            The host's ip address.
        
        host_name : `str`
            The hosts' name.
        
        port : `int`
            Port to connect to the host.
        
        protocol : `int`
            Protocol type.
        """
        self = object.__new__(cls)
        self.family = family
        self.flags = flags
        self.host = host
        self.host_name = host_name
        self.port = port
        self.protocol = protocol
        return self
    
    
    @classmethod
    def from_ip(cls, host, port, family):
        """
        Creates a host info instance from the given parameters.
        Used when `host` is an ip address.
        
        Parameters
        ----------
        host : `str`
            The host's ip address.
        port : `int`
            Port to connect to the host.
        family : `AddressFamily`
            Address family.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.family = family
        self.flags = 0
        self.host = host
        self.host_name = host
        self.port = port
        self.protocol = 0
        return self
    
    
    @classmethod
    def from_address_info(cls, host_name, address_info):
        """
        Creates a host info instance from the given parameters.
        
        Parameters
        ----------
        host_name : `str`
            The host's name.
        
        address_info : `(AddressFamily, SocketKind, int, str, (str, int) | (str, int, int, int))`
            An address info returned by `get_address_info`.
        
        Returns
        -------
        self : `instance<cls>`
        """
        address = address_info[4]
        
        # Construct
        self = object.__new__(cls)
        self.family = address_info[0]
        self.flags = ADDRESS_INFO_NUMERIC_HOST
        self.host = address[0]
        self.host_name = host_name
        self.port = address[1]
        self.protocol = address_info[2]
        
        return self
    
    
    def __repr__(self):
        """Returns the host info's representation."""
        repr_parts = ['<', type(self).__name__]
        
        # family
        family = self.family
        if family:
            repr_parts.append(' family = ')
            repr_parts.append(repr(family))
            
            field_added = True
        else:
            field_added = False
        
        # flags
        flags = self.flags
        if flags:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' flags = ')
            repr_parts.append(repr(flags))
        
        # host
        if field_added:
            repr_parts.append(',')
        
        host = self.host
        repr_parts.append(' host = ')
        repr_parts.append(repr(host))
        
        # host_name
        host_name = self.host_name
        if (host != host_name):
            repr_parts.append(', host_name = ')
            repr_parts.append(repr(host_name))
        
        # port
        repr_parts.append(', port = ')
        repr_parts.append(repr(self.port))
        
        # protocol
        protocol = self.protocol
        if protocol:
            repr_parts.append(', protocol = ')
            repr_parts.append(repr(protocol))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two host info's are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # family
        if self.family != other.family:
            return False
        
        # flags
        if self.flags != other.flags:
            return False
        
        # host
        if self.host != other.host:
            return False
        
        # host_name
        if self.host_name != other.host_name:
            return False
        
        # port
        if self.port != other.port:
            return False
        
        # protocol
        if self.protocol != other.protocol:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the host info's hash value."""
        hash_value = 0
        
        # family
        hash_value ^= hash(self.family)
        
        # flags
        hash_value ^= hash(self.flags)
        
        # host
        host = self.host
        hash_value ^= hash(host)
        
        # host_name
        host_name = self.host_name
        if (host != host_name):        
            hash_value ^= hash(host_name)
        
        # port
        hash_value ^= self.port << 17
        
        # protocol
        hash_value ^= hash(self.protocol)
        
        return hash_value



__all__ = ()

from ...utils import copy_docs, copy_func

from ..quoting import quote, unquote

from .url_part_base_parsed import URLPartBaseParsed


encode = lambda value: quote(value, safe = '@:', protected = '/')
decode = unquote


def normalize_path_parts(parts):
    """
    Normalises the given parts.
    
    Parameters
    ----------
    parts : `list<str>`
        The parts to normalize.
    
    Returns
    -------
    parts : `None | list<str>`
        The normalized parts.
    """
    if parts is None:
        return None

    parts_length = len(parts)
    deleted_count = 0
    
    # Remove single dot parts. If the last is dot set it as empty string instead.
    index = 0
    while index < parts_length:
        part = parts[index]
        if part != '.':
            index += 1
            continue
        
        if index and index == parts_length - 1:
            parts[index] = ''
            index += 1
        else:
            del parts[index]
            parts_length -= 1
            deleted_count += 1
    
    # Remove double dot parts.
    to_delete = 0
    
    for index in reversed(range(parts_length)):
        part = parts[index]
        if part == '..':
            to_delete += 1
            continue
        
        if not to_delete:
            continue
        
        to_delete -= 1
        del parts[index]
        del parts[index]
        deleted_count += 2
        parts_length -= 2
        
    
    while to_delete:
        to_delete -= 1
        del parts[0]
        deleted_count += 1
        parts_length -= 1
        
    if deleted_count and (parts_length == 1) and (not parts[0]):
        del parts[0]
        parts_length = 0
    
    if (parts_length == 0):
        parts = None
    
    return parts


class URLPath(URLPartBaseParsed):
    """
    Represents a url's path.
    
    Attributes
    ----------
    _decoded : `None | str`
        Decoded value.
    
    _encoded : `None | str`
        Encoded value.
    
    _flags : `int`
        Bitwise flags representing how the object is constructed.
    
    parsed : `None | tuple<str>`
        The represented path split to parts.
    """
    __slots__ = ()
    
    
    _encode = staticmethod(copy_docs(URLPartBaseParsed._encode)(copy_func(encode)))
    _decode = staticmethod(copy_docs(URLPartBaseParsed._decode)(copy_func(decode)))
    
    
    @staticmethod
    @copy_docs(URLPartBaseParsed._parse)
    def _parse(value, encoded):
        if (value is not None):
            if not value:
                return None
            
            if value.startswith('/'):
                value = value[1:]
            
            parts = [(decode(part) if encoded else part) for part in value.split('/')]
            if not parts[0]:
                del parts[0]
            
            parts = normalize_path_parts(parts)
            if (parts is not None):
                return tuple(parts)
    
    
    @staticmethod
    @copy_docs(URLPartBaseParsed._serialize)
    def _serialize(value, encoded):
        if (value is not None):
            built = []
            
            if not value:
                built.append('/')
            
            else:
                for part in value:
                    if encoded:
                        part = encode(part)
                
                    built.append('/')
                    built.append(part)
            
            return ''.join(built)



__all__ = ()

from ...utils import RichAttributeErrorBaseType

from .url_part_base import URLPartBase


class URLUserInfo(RichAttributeErrorBaseType):
    """
    Represents the user information of an url.
    
    Attributes
    ----------
    _name : `None | URLPartBase`
        The user's name if any.
    
    _password : `None | URLPartBase`
        The user's password if any.
    """
    __slots__ = ('_name', '_password')
    
    
    def __new__(cls):
        """
        Use either ``.create_from_decoded`` or ``.create_from_encoded`` instead.
        
        Raises
        ------
        RuntimeError
        """
        raise RuntimeError(
            'Use either `.create_from_decoded` or `.create_from_encoded` instead.'
        )
    
    
    def __repr__(self):
        """Returns the url path's representation."""
        repr_parts = ['<', type(self).__name__]
        
        field_added = False
        
        # name_decoded
        name_decoded = self.name_decoded
        if (name_decoded is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' name_decoded = ')
            repr_parts.append(repr(name_decoded))
        
        # password_decoded
        password_decoded = self.password_decoded
        if (password_decoded is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' password_decoded = ')
            repr_parts.append(repr(password_decoded))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two url parts are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # name
        if self._name != other._name:
            return False
        
        # password
        if self._password != other._password:
            return False
        
        return True
    
    
    @classmethod
    def create_from_decoded(cls, user_name, user_password):
        """
        Creates a new url user info from decoded values.
        
        Parameters
        ----------
        user_name : `None | str`
            The user's name.
        
        user_password : `None | str`
            The user's password.
        
        Returns
        -------
        self : `instance<cls>`
        """
        if user_name is None:
            name = None
        else:
            name = URLPartBase.create_from_decoded(user_name)
        
        if user_password is None:
            password = None
        else:
            password = URLPartBase.create_from_decoded(user_password)
        
        self = object.__new__(cls)
        self._name = name
        self._password = password
        return self
    
    
    @classmethod
    def create_from_encoded(cls, user_name, user_password):
        """
        Creates a new url user info from encoded values.
        
        Parameters
        ----------
        user_name : `None | str`
            The user's name.
        
        user_password : `None | str`
            The user's password.
        
        Returns
        -------
        self : `instance<cls>`
        """
        if user_name is None:
            name = None
        else:
            name = URLPartBase.create_from_encoded(user_name)
        
        if user_password is None:
            password = None
        else:
            password = URLPartBase.create_from_encoded(user_password)
        
        self = object.__new__(cls)
        self._name = name
        self._password = password
        return self
    
    
    @property
    def name_decoded(self):
        """
        Returns the user info's decoded name.
        
        Returns
        -------
        name : `None | str`
        """
        name = self._name
        if (name is not None):
            return name.decoded
    
    
    @property
    def password_decoded(self):
        """
        Returns the user info's decoded password.
        
        Returns
        -------
        name : `None | str`
        """
        password = self._password
        if (password is not None):
            return password.decoded

    
    @property
    def name_encoded(self):
        """
        Returns the user info's encoded name.
        
        Returns
        -------
        name : `None | str`
        """
        name = self._name
        if (name is not None):
            return name.encoded
    
    
    @property
    def password_encoded(self):
        """
        Returns the user info's encoded password.
        
        Returns
        -------
        name : `None | str`
        """
        password = self._password
        if (password is not None):
            return password.encoded



from .frame_proxy_base import *
from .frame_proxy_frame import *
from .frame_proxy_traceback import *
from .frame_proxy_virtual import *
from .frame_proxying import *


# Empty all, we just wanna pass the references down and not outside
__all__ = ()



__all__ = ('get_matching_variable_names_with_attribute',)

from .helpers_familiarity import SUGGESTIONS_MAX
from .helpers_variables import _iter_unique_variables


def get_matching_variable_names_with_attribute(frame, name):
    """
    Gets all the familiar attribute names of the variable to the given one.
    
    Parameters
    ----------
    frame : `None | FrameProxyBase`
        Frame to get variable names from.
    name : `str`
        Name to check for.
    
    Returns
    -------
    matches : `None | list<str>`
    """
    matches = []
    
    for variable_name, variable_value in _iter_unique_variables(frame):
        if hasattr(variable_value, name):
            matches.append(variable_name)
    
    if not matches:
        return None
    
    del matches[SUGGESTIONS_MAX:]
    matches.sort()
    return matches



__all__ = ('FileInfo',)

from os import stat as get_stats

from .cache_constants import FILE_INFO_CACHE, LINE_CACHE_SESSIONS


class FileInfo:
    """
    Represents a file.
    
    Attributes
    ----------
    alive : `bool`
        Whether the file's source is alive.
    console : `bool`
        Whether the file's source is from console.
    file_modification_time : `float`
        When the file was last modified. Using system time. Used to identify file modifications.
    file_name : `str`
        The line's owner file's name.
    file_size : `int`
        The file's size. Used to identify file modifications.
    lines : `list<str>`
        The file's lines.
    """
    __slots__ = ('__weakref__', 'alive', 'console', 'file_modification_time', 'file_name', 'file_size', 'lines')
    
    def __new__(cls, file_name, file_size, file_modification_time, lines, alive, console):
        """
        Creates a new file info.
        
        Parameters
        ----------
        file_name : `str`
            The line's owner file's name.
        file_size : `int`
            The file's size.
        file_modification_time : `float`
            When the file was last modified.
        lines : `list<str>`
            The file's lines.
        alive : `bool`
            Whether the file's source is alive.
        console : `bool`
            Whether the file's source is from console.
        """
        self = object.__new__(cls)
        self.alive = alive
        self.console = console
        self.file_modification_time = file_modification_time
        self.file_name = file_name
        self.file_size = file_size
        self.lines = lines
        return self
    
    
    def __repr__(self):
        """Returns the file info's representation."""
        repr_parts = ['<', type(self).__name__]
        
        repr_parts.append(', alive = ')
        repr_parts.append(repr(self.alive))
        
        repr_parts.append(', console = ')
        repr_parts.append(repr(self.console))
        
        repr_parts.append(' file_modification_time = ')
        repr_parts.append(repr(self.file_modification_time))
        
        repr_parts.append(', file_name = ')
        repr_parts.append(repr(self.file_name))
        
        repr_parts.append(', file_size = ')
        repr_parts.append(repr(self.file_size))
        
        repr_parts.append(', lines = ')
        repr_parts.append(repr(self.lines))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def get_line(self, index):
        """
        Gets the line of the file.
        
        Parameters
        ----------
        index : `int`
            The line's index.
        
        Returns
        -------
        hit : `bool`
        line : `str`
        """
        if index < 0:
            return False, ''
        
        lines = self.lines
        if index >= len(lines):
            return False, ''
        
        return True, lines[index]


def get_file_info(file_name):
    """
    Gets file info for the given file name.
    
    Parameters
    ----------
    file_name : `str`
        The file's name.
    
    Returns
    -------
    file_info : ``FileInfo``
    """
    # Pull from cache
    file_info = FILE_INFO_CACHE.get(file_name, None)
    
    if (file_info is None):
        # Try create
        try:
            stats = get_stats(file_name)
        except OSError:
            file_info = None
        else:
            try:
                with open(file_name, 'r', encoding = 'utf-8', errors = 'replace') as file:
                    lines = file.read().splitlines()
            except OSError:
                file_info = None
            else:
                file_info = FileInfo(file_name, stats.st_size, stats.st_mtime, lines, True, False)
    
    # Could not create full instance, create dummy.
    if file_info is None:
        file_info = FileInfo(file_name, 0, 0.0, [], False, False)
    
    # Put file to temporary cache
    FILE_INFO_CACHE[file_name] = file_info
    for line_cache_context in LINE_CACHE_SESSIONS:
        line_cache_context.memorized_file_infos.add(file_info)
    
    return file_info




__all__ = ()

from datetime import datetime as DateTime
from math import isinf, isnan

from ...utils import MultiValueDictionary, copy_docs, copy_func

from ..quoting import quote, unquote

from .constants import NoneType
from .url_part_base_parsed import URLPartBaseParsed


encode = lambda value: quote(value, safe = '/?:@=', query_string = True)
decode = unquote


def parse_query(query_string, encoded):
    """
    Parses a query string.
    
    Parameters
    ----------
    query_string : `None | str`
        Query string to parse if any.
    
    encoded : `bool`
        Whether the query is encoded.
    
    Returns
    -------
    query : `None | MultiValueDictionary<str, str>`
    """
    if query_string is None:
        return
    
    query = None
    value_length = len(query_string)
    item_end_index = 0
    
    while item_end_index < value_length:
        item_start_index = item_end_index
        
        # Find item end
        item_end_index = query_string.find('&', item_start_index)
        if item_end_index == -1:
            item_end_index = len(query_string)
        
        if item_start_index != item_end_index:
            # Find item middle
            item_middle_index = query_string.find('=', item_start_index, item_end_index)
            if item_middle_index == -1:
                query_key = query_string[item_start_index : item_end_index]
                query_value = ''
            
            else:
                query_key = query_string[item_start_index : item_middle_index]
                query_value = query_string[item_middle_index + 1 : item_end_index]
            
            if encoded:
                query_key = decode(query_key)
                query_value = decode(query_value)
            
            if query is None:
                query = MultiValueDictionary()
            
            query[query_key] = query_value
        
        item_end_index += 1
        continue
    
    return query



def normalize_query(query):
    """
    Normalises the given query.
    
    Parameters
    ----------
    query : `None | str | dict<str, str | int | bool | None | float | DateTime | (list | set)<...>> | (list | set)<(str, ...)>`
        The query to normalize.
    
    Returns
    -------
    normalized_query : `None | MultiValueDictionary<str, str>`
    
    Raises
    ------
    TypeError
    ValueError
    """
    if query is None:
        normalized_query = None
    
    elif isinstance(query, dict):
        normalized_query = _normalize_query_dict(None, query)
    
    elif isinstance(query, (list, set)):
        normalized_query = _normalize_query_list(None, query)
    
    elif isinstance(query, str):
        normalized_query = parse_query(query, False)
    
    elif isinstance(query, (bytes, bytearray, memoryview)):
        raise TypeError(
            f'Query type cannot be `bytes`, `bytearray` and `memoryview`, got '
            f'{type(query).__name__}; {query!r}.'
        )
    
    else:
        raise TypeError(
            f'Invalid query type: {type(query).__name__}; {query!r}.'
        )
    
    return normalized_query



def _normalize_query_dict(normalized_query, query_dict):
    """
    Normalises a query dict.
    
    Parameters
    ----------
    normalized_query : `None | MultiValueDictionary<str, str>`
        The already normalised query.
    
    query_dict : `dict<str, str | int | bool | None | float | DateTime | (list | set)<...>>`
        The query dict to normalize.
    
    Returns
    -------
    normalized_query : `None | MultiValueDictionary<str, str>`
    
    Raises
    ------
    TypeError
    ValueError
    """
    for query_key, query_value in query_dict.items():
        normalized_query = _normalize_query_item(normalized_query, query_key, query_value)
    
    return normalized_query


def _normalize_query_list(normalized_query, query_list):
    """
    Normalises a query list..
    
    Parameters
    ----------
    normalized_query : `None | MultiValueDictionary<str, str>`
        The already normalised query.
    
    query_dict : `(list | set)<(str, str | int | bool | None | float | DateTime | (list | set)<...>)>`
        The query list to normalize.
    
    Returns
    -------
    normalized_query : `None | MultiValueDictionary<str, str>`
    
    Raises
    ------
    TypeError
    ValueError
    """
    for query_key, query_value in query_list:
        normalized_query = _normalize_query_item(normalized_query, query_key, query_value)
    
    return normalized_query


def _normalize_query_item(normalized_query, query_key, query_value):
    """
    Normalises a query item (key-value pair).
    
    Parameters
    ----------
    normalized_query : `None | MultiValueDictionary<str, str>`
        The already normalised query.
    
    query_key : `str`
        Query key.
    
    query_value : `str | int | bool | None | float | DateTime | (list | set)<...>`
        Query value.
    
    Returns
    -------
    normalized_query : `None | MultiValueDictionary<str, str>`
    
    Raises
    ------
    TypeError
    ValueError
    """
    if isinstance(query_value, str):
        pass
    
    elif isinstance(query_value, bool):
        query_value = 'true' if query_value else 'false'
    
    elif isinstance(query_value, int):
        query_value = str(query_value)
    
    elif isinstance(query_value, NoneType):
        query_value = 'null'
    
    elif isinstance(query_value, DateTime):
        query_value = query_value.isoformat()
    
    elif isinstance(query_value, float):
        if isinf(query_value):
             raise ValueError('`inf` is not a supported query string parameter value.')
        
        if isnan(query_value):
            raise ValueError('`nan` is not a supported query string parameter value.')
        
        query_value = str(query_value)
    
    elif isinstance(query_value, (list, tuple, set)):
        return _normalize_query_item_list(normalized_query, query_key, query_value)
    
    else:
        raise TypeError(
            f'Unexpected value type received when serializing query string, got '
            f'{type(query_value).__name__}; {query_value!r}.'
        )
    
    if (not query_key) and (not query_value):
        return normalized_query
    
    if normalized_query is None:
        normalized_query = MultiValueDictionary()
    
    normalized_query[query_key] = query_value
    return normalized_query


def _normalize_query_item_list(normalized_query, query_key, query_value_list):
    """
    Normalises a query item where the value is a list of more values.
    
    Parameters
    ----------
    normalized_query : `None | MultiValueDictionary<str, str>`
        The already normalised query.
    
    query_key : `str`
        Query key.
    
    query_value : `(list | set)<tr | int | bool | None | float | DateTime | ...>`
        Query value.
    
    Returns
    -------
    normalized_query : `None | MultiValueDictionary<str, str>`
    
    Raises
    ------
    TypeError
    ValueError
    """
    for query_value in query_value_list:
        normalized_query = _normalize_query_item(normalized_query, query_key, query_value)
    
    return normalized_query


def serialize_query(query, encoded):
    """
    Serialises the given query.
    
    Parameters
    ----------
    query : `None | MultiValueDictionary<str, str>`
        The query to serialise.
    
    encoded : `bool
        Whether the output should be encoded.
    
    Returns
    -------
    query_string : `None | str`
    """
    if (query is None) or (not query):
        return None
    
    query_string_parts = []
    
    # Sort the keys to always get the same output.
    for query_key in sorted(query.keys()):
        if encoded:
            query_key_processed = encode(query_key)
        else:
            query_key_processed = query_key
        
        for query_value in query.get_all(query_key, ()):
            if encoded:
                query_value = encode(query_value)
            
            if query_string_parts:
                query_string_parts.append('&')
            
            query_string_parts.append(query_key_processed)
            query_string_parts.append('=')
            query_string_parts.append(query_value)
    
    return ''.join(query_string_parts)


class URLQuery(URLPartBaseParsed):
    """
    Represents a url's query.
    
    Attributes
    ----------
    _decoded : `None | str`
        Decoded value.
    
    _encoded : `None | str`
        Encoded value.
    
    _flags : `int`
        Bitwise flags representing how the object is constructed.
    
    parsed : `None | tuple<str>`
        The represented path split to parts.
    """
    __slots__ = ()
    
    
    _encode = staticmethod(copy_docs(URLPartBaseParsed._encode)(copy_func(encode)))
    _decode = staticmethod(copy_docs(URLPartBaseParsed._decode)(copy_func(decode)))
    
    
    _parse = staticmethod(copy_docs(URLPartBaseParsed._parse)(copy_func(parse_query)))
    _serialize = staticmethod(copy_docs(URLPartBaseParsed._serialize)(copy_func(serialize_query)))



from .exception_representation import *
from .expression_parsing import *
from .frame_proxy import *

from .formatters import *
from .frame_group import *
from .frame_grouping import *
from .frame_ignoring import *
from .rendering import *
from .repeat_strategies import *
from .trace import *


__all__ = (
    *exception_representation.__all__,
    *expression_parsing.__all__,
    *frame_proxy.__all__,
    
    *formatters.__all__,
    *frame_group.__all__,
    *frame_grouping.__all__,
    *frame_ignoring.__all__,
    *rendering.__all__,
    *repeat_strategies.__all__,
    *trace.__all__,
)



__all__ = ()

from codecs import lookup as lookup_encoding
from http import HTTPStatus
from http.cookies import CookieError, SimpleCookie
from warnings import warn

from ..utils import RichAttributeErrorBaseType, from_json
from ..web_common import parse_content_type
from ..web_common.headers import CONTENT_TYPE, METHOD_CONNECT, METHOD_HEAD, SET_COOKIE

from .constants import JSON_RE

try:
    from cchardet import detect as detect_encoding
except ImportError:
    try:
        from chardet import detect as detect_encoding
    except ImportError as err:
        detect_encoding = None


class ClientResponse(RichAttributeErrorBaseType):
    """
    Http response class used by ``HTTPClient``.
    
    Attributes
    ----------
    _released : `bool`
        Whether the connection is released.
    
    body : `None | bytes`
        The received response body. Set as `None` if the response body is not yet received, or if it is empty.
        Not set as non-`None` if the payload was streamed.
    
    closed : `bool`
        Whether the response is closed.
    
    connection : `None | Connection`
        Connection used to receive the request response.
        Set as `None` if the response is ``.close``-d or ``.release``-d.
    
    cookies : `http.cookies.SimpleCookie`
        Received cookies with the response.
    
    history : `None | tuple<ClientResponse>`
        Response history. Set as `tuple` of responses from outside.
    
    loop : ``EventThread``
        The event loop, trough what the request is executed.
    
    method : `str`
        Method of the respective request.
    
    payload_stream : `None | PayloadStream`
        Future used to retrieve the response's body.
        It's result is set, when the respective protocol's reader task finished.
    
    raw_message : `None | RawResponseMessage`
        Raw received http response.
    
    url : ``URL``
        The requested url.
    
    write_body_task : `None | Task<ClientRequest.write_body>`
        Payload writer task of the respective request.
    """
    __slots__ = (
        '_released', 'body', 'closed', 'connection', 'cookies', 'history', 'loop', 'method', 'payload_stream',
        'raw_message', 'url', 'write_body_task' 
    )
       
    def __new__(cls, request, connection):
        """
        Crates a new client response from the given request and connection.
        
        Parameters
        ----------
        request : ``ClientRequest``
            The respective request.
        
        connection : ``Connection``
            The connection used to send the request and receive the response.
        """
        self = object.__new__(cls)
        
        self._released = False
        self.body = None
        self.closed = False
        self.connection = connection
        self.cookies = SimpleCookie()
        self.history = None
        self.loop = request.loop
        self.method = request.method
        self.payload_stream = None
        self.raw_message = None
        self.url = request.original_url
        self.write_body_task = request.write_body_task
        
        return self
    
    
    def __repr__(self):
        """Returns the response's representation."""
        repr_parts = ['<', type(self).__name__]
        
        # url
        repr_parts.append(' url = ')
        repr_parts.append(str(self.url))
        
        # status & reason
        repr_parts.append(', response = ')
        repr_parts.append(repr(f'{self.status!s} {self.reason!s}'))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __del__(self):
        """releases the response if not yet closed."""
        if self.closed:
            return
        
        self._release_connection()
    
    
    @property
    def headers(self):
        """
        Headers of the response. Set when the http response is successfully received.
        
        Returns
        --------
        headers : `None | IgnoreCaseMultiValueDictionary<str, str>`
        """
        message = self.raw_message
        if (message is None):
            return None
        
        return message.headers
    
    
    @property
    def reason(self):
        """
        Returns the server response reason.
        
        Returns
        -------
        reason : `None | str`
        """
        message = self.raw_message
        if (message is None):
            return None
        
        reason = message.reason
        if (reason is not None):
            return reason
        
        return HTTPStatus(message.status).phrase
    
    
    @property
    def status(self):
        """
        Received status code. Returns `0` by default.
        
        Returns
        -------
        status : `int`
        """
        message = self.raw_message
        if (message is None):
            return 0
        
        return message.status
    
    
    async def start_processing(self):
        """
        Starts response processing.
        
        This method is a coroutine.
        """
        try:
            protocol = self.connection.protocol
            self.raw_message = message = await protocol.read_http_response()
            
            if self.method == METHOD_CONNECT:            
                # Do not read anything if we are doing a connect request.
                payload_stream = None
            
            else:
                if self.method == METHOD_HEAD:
                    payload_reader = None
                else:
                    payload_reader = protocol.get_payload_reader_task(message)
                
                if (payload_reader is None):
                    payload_stream = None
                    self._response_eof(None)
                else:
                    payload_stream = protocol.set_payload_reader(payload_reader)
                    payload_stream.add_done_callback(self._response_eof)
                    protocol.handle_payload_stream_abortion()
            
            self.payload_stream = payload_stream
            
            # cookies
            headers = message.headers
            if (headers is not None):
                for header in headers.get_all(SET_COOKIE, ()):
                    try:
                        self.cookies.load(header)
                    except CookieError: # so sad
                        pass
        except:
            self.close()
            raise
    
    
    def _response_eof(self, payload_stream):
        """
        Future callback added to the payload waiter future, to release the used connection.
        
        Parameters
        ----------
        payload_stream : `None | PayloadStream`
            The respective payload stream.
        """
        if self.closed:
            return
        
        connection = self.connection
        if (connection is not None):
            # WebSocket, protocol could be `None`, because connection could be detached.
            if (connection.protocol is not None) and self.raw_message.upgraded:
                return
            
            self._released = True
            self._release_connection()
        
        self.closed = True
        self._clean_up_writer()
    
    
    def _release_connection(self):
        """
        Releases the response's connection.
        
        If the connection type is `close`, closes the protocol as well.
        """
        connection = self.connection
        if connection is None:
            return
        
        raw_message = self.raw_message
        if raw_message is None:
            keep_alive = None
        else:
            keep_alive = raw_message.keep_alive
        
        self.connection = None
        connection.release(keep_alive)
    
    
    def _notify_content(self):
        """
        Called when response reading is cancelled or released.
        Sets `ConnectionError` to the respective protocol if the payload is still reading.
        """
        payload_stream = self.payload_stream
        if (payload_stream is not None):
            connection = self.connection
            if (connection is not None):
                protocol = connection.protocol
                if (protocol is not None):
                    protocol.set_exception(ConnectionError('Connection closed.'))
        
        self._released = True
    
    
    def _clean_up_writer(self):
        """
        Cancels the writer task of the respective request. Called when the response is cancelled or released, or if
        reading the whole response is done.
        """
        write_body_task = self.write_body_task
        if (write_body_task is not None):
            self.write_body_task = None
            write_body_task.cancel()
    
    
    async def read(self):
        """
        Reads the response's body.
        
        This method is a coroutine.
        
        Returns
        -------
        body : `None | bytes`
        """
        payload_stream = self.payload_stream
        if (payload_stream is None):
            body = self.body
        else:
            try:
                self.body = body = await payload_stream
            finally:
                self.payload_stream = None
        
        return body
    
    
    def get_encoding(self):
        """
        Gets the encoding of the response's body.
        
        Returns
        -------
        encoding : `str`
            Defaults to `'utf-8'`.
        """
        headers = self.headers
        if headers is None:
            return 'utf-8'
        
        content_type, content_type_parsing_error = parse_content_type(headers.get(CONTENT_TYPE, ''))
        
        encoding = content_type.get_parameter('charset', None)
        if (encoding is not None):
            encoding = encoding.casefold()
            try:
                lookup_encoding(encoding)
            except LookupError:
                pass
            else:
                return encoding
        
        # RFC 7159 states that the default encoding is utf-8.
        if (content_type.type == 'application' and content_type.sub_type in ('json', 'rdap')):
            return 'utf-8'
        
        # If we cannot detect encoding leave
        if (detect_encoding is None):
            return 'utf-8'
        
        # Can we detect encoding from anything even?
        body = self.body
        if (body is None):
            return 'utf-8'
        
        encoding = detect_encoding(body)['encoding']
        if encoding is None:
            encoding = 'utf-8'
        
        return encoding.casefold()
    
    
    async def text(self, *deprecated, encoding = None, errors = 'strict'):
        """
        Loads the response's content as text.
        
        This method is a coroutine.
        
        Parameters
        ----------
        encoding : `None`, `str` = `None`, Optional (Keyword only)
            If no encoding is given, then detects it from the payload.
        
        errors : `str` = `'strict'`, Optional (Keyword only)
            May be given to set a different error handling scheme. The default `errors` value is `'strict'`, meaning
            that encoding errors raise a `UnicodeError`. Other possible values are `'ignore'`, `'replace'`,
            `'xmlcharrefreplace'`, `'backslashreplace'` and any other name registered via `codecs.register_error()`.
        
        Returns
        -------
        text : `str`
        """
        # deprecated
        deprecated_length = len(deprecated)
        if deprecated_length:
            warn(
                (
                    f'The `encoding` and `errors` parameters of '
                    f'`{type(self).__name__}.text` are moved to be keyword only. '
                    f'Support for positional is deprecated and will be removed in 2025 August.'
                ),
                FutureWarning,
                stacklevel = 2,
            )
            
            encoding = deprecated[0]
            
            if deprecated_length > 1:
                errors = deprecated[1]
        
        
        body = await self.read()
        if body is None:
            return
        
        if encoding is None:
            encoding = self.get_encoding()
        
        return body.decode(encoding, errors)
    
    
    async def json(self, *deprecated, content_type = None, encoding = None, loader = from_json):
        """
        Loads the response's content as a json.
        
        This method is a coroutine.
        
        Parameters
        ----------
        content_type : `None | str` = `None`, Optional (Keyword only)
            Content type to use instead of the default one.
            Pass it as empty string to disable the check.
        
        encoding : `None | str` = `None`, Optional (Keyword only)
            Encoding to use instead of the response's.
            If not given then will use the response's own encoding.
        
        loader : `callable` = ``from_json``, Optional (Keyword only)
            Json loader. Defaults to json.loads`.
        
        Returns
        -------
        json : `object`
        
        Raises
        ------
        TypeError
            If the response's mime_type do not match.
        """
        # deprecated
        deprecated_length = len(deprecated)
        if deprecated_length:
            warn(
                (
                    f'The `encoding`, `loader` and `content_type` parameters of '
                    f'`{type(self).__name__}.json` are moved to be keyword only. '
                    f'Support for positional is deprecated and will be removed in 2025 August.'
                ),
                FutureWarning,
                stacklevel = 2,
            )
            
            encoding = deprecated[0]
            
            if deprecated_length > 1:
                loader = deprecated[1]
            
            if deprecated_length > 2:
                content_type = deprecated[2]
        
        # body
        body = await self.read()
        if body is None:
            return None
        
        
        # content_type
        while True:
            # At this point we should have headers, but in tests we have only `.body`.
            headers = self.headers
            if headers is None:
                break
            
            received_content_type = headers.get(CONTENT_TYPE, '').casefold()
            
            if content_type is None:
                if (JSON_RE.match(received_content_type) is not None):
                    break
            
            else:
                if (content_type in received_content_type):
                    break
            
            raise TypeError(
                f'Attempt to decode JSON with unexpected mime_type: {received_content_type!r}.'
            )
        
        # encoding
        if encoding is None:
            encoding = self.get_encoding()
        
        return loader(body.decode(encoding))
    
    
    def close(self):
        """
        Closes the response and it's connection. The used connection will not be reused after.
        """
        if not self._released:
            self._notify_content()
        
        if self.closed:
            return
        
        self.closed = True
        
        connection = self.connection
        if (connection is not None):
            self.connection = None
            connection.close()
        
        self._clean_up_writer()
    
    
    def release(self):
        """
        Releases the response and it's connection. The used connection might be reused after.
        """
        if not self._released:
            self._notify_content()
        
        if self.closed:
            return
        
        self.closed = True
        
        self._release_connection()
        self._clean_up_writer()
    
    
    @property
    def payload_waiter(self):
        """
        Deprecated and will be removed in 2025 October. Please use `.payload_stream` instead.
        """
        warn(
            (
                f'`{type(self).__name__}.payload_waiter` is deprecated and will be removed in 2025 October. '
                f'Please use `.payload_stream` instead. Note that it is a different type.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
    
        return self.payload_stream



__all__ = ()

from http.cookies import SimpleCookie
from warnings import warn

from ..core import LOOP_TIME
from ..utils import RichAttributeErrorBaseType

from .connection import Connection
from .protocol_basket import ProtocolBasket


class ConnectorBase(RichAttributeErrorBaseType):
    """
    Base connector class.
    
    Attributes
    ----------
    clean_up_handle : `None | TimerWeakHandle`
        Weak handle, which cleans up the timed out connections of the connector.
    
    closed : `bool`
        Whether the connector is closed.
        
    cookies : `http.cookies.SimpleCookie`
        Cookies of the connection.
    
    force_close : `bool`
        Whether after each request (and between redirects) the connections should be closed.
    
    loop : ``EventThread``
        The event loop to what the connector is bound to.
    
    protocols_by_host : `dict<ConnectionKey, ConnectionBasket>`
        Protocols for each host.
    
    Notes
    -----
    Connectors support weakreferencing.
    """
    __slots__ = (
        '__weakref__', 'clean_up_handle', 'closed', 'cookies', 'force_close', 'loop', 'protocols_by_host'
    )
    
    def __new__(cls, loop, *deprecated, force_close = False):
        """
        Creates a new connector bound to the given loop.
        
        Parameters
        ----------
        loop : ``EventThread``
            The event loop to what the connector is bound to.
        
        force_close : `bool` = `False`, Optional (Keyword only)
            Whether after each request (and between redirects) the connections should be closed. Defaults to `False`.
        """
        # deprecated
        deprecated_length = len(deprecated)
        if deprecated_length:
            warn(
                (
                    f'The `force_close` parameter in `{cls.__name__}.__new__` is moved ot be keyword only. '
                    f'Support for positional is deprecated and will be removed in 2025 August.'
                ),
                FutureWarning,
                stacklevel = 2,
            )
            
            force_close = deprecated[0]
        
        self = object.__new__(cls)
        self.protocols_by_host = {}
        self.clean_up_handle = None
        self.closed = False
        self.cookies = SimpleCookie()
        self.force_close = force_close
        self.loop = loop
        return self
    
    
    def __del__(self):
        """
        Closes the connector if not yet closed.
        """
        self.close()
    
    
    def _clean_up(self):
        """
        Cleans ups the expired connections of the connector.
        """
        handle = self.clean_up_handle
        if (handle is not None):
            self.clean_up_handle = None
            handle.cancel()
        
        # Clean up unused transports.
        protocols_by_host = self.protocols_by_host
        if not protocols_by_host:
            return
        
        now = LOOP_TIME()
        to_remove_keys = None
        
        for protocol_basket in protocols_by_host.values():
            protocol_basket.clean_up_expired_protocols(now)
            if protocol_basket:
                continue
            
            if to_remove_keys is None:
                to_remove_keys = []
            
            to_remove_keys.append(protocol_basket.connection_key)
        
        if (to_remove_keys is not None):
            for key in to_remove_keys:
                del protocols_by_host[key]
        
        closest_expiration = self.get_closest_expiration()
        if closest_expiration != -1:
            self.clean_up_handle = self.loop.call_at_weak(closest_expiration, self._clean_up)
    
    
    def close(self):
        """
        Closes the connector and it's connections.
        """
        if self.closed:
            return
        
        self.closed = True
        
        protocols_by_host = self.protocols_by_host
        try:
            if not self.loop.running:
                return
            
            for protocol_basket in protocols_by_host.values():
                protocol_basket.close()
            
        finally:
            protocols_by_host.clear()
    
    
    async def connect(self, request):
        """
        Gets connection from connection pool or creates a new one.
        
        This method is a coroutine.
        
        Parameters
        ----------
        request : ``ClientRequest``
            Respective request, which requires a connection.
        
        Returns
        -------
        connection : ``Connection``
            The created connection.
        
        Raises
        ------
        ConnectionError
            Connector closed.
        """
        key = request.connection_key
        
        protocol, performed_requests = self.pop_available_protocol(key)
        if protocol is None:
            protocol = await self.create_connection(request)
            if self.closed:
                protocol.close()
                raise ConnectionError('Connector is closed.')
        
        self.add_used_protocol(key, protocol)
        return Connection(self, key, protocol, performed_requests)
    
    
    def pop_available_protocol(self, key):
        """
        Gets a protocol for the given connection key.
        
        Parameters
        ----------
        key : ``ConnectionKey``
            A key which contains information about the host.

        Returns
        -------
        protocol : `None | AbstractProtocolBase`
            Protocol connected to the respective host. Defaults to `None` if there is not any.
        
        performed_requests : `int`
            The amount of performed requests on the protocol.
        """
        protocols_by_host = self.protocols_by_host
        try:
            protocol_basket = protocols_by_host[key]
        except KeyError:
            return None, 0
        
        protocol_and_performed_requests = protocol_basket.pop_available_protocol(LOOP_TIME())
        if not protocol_basket:
            del protocols_by_host[key]
        
        return protocol_and_performed_requests
    
    
    def add_used_protocol(self, key, protocol):
        """
        Adds a protocol as used.
        
        Parameters
        ----------
        key : ``ConnectionKey``
            A key that contains information about the host.
        
        protocol : ``AbstractProtocolBase``
            The protocol to add.
        """
        protocols_by_host = self.protocols_by_host
        try:
            protocol_basket = protocols_by_host[key]
        except KeyError:
            protocol_basket = ProtocolBasket(key)
            protocols_by_host[key] = protocol_basket
        
        protocol_basket.add_used_protocol(protocol)
    
    
    def add_available_protocol(self, key, protocol, keep_alive_timeout, performed_requests):
        """
        Adds a protocol as available.
        
        Parameters
        ----------
        key : ``ConnectionKey``
            A key that contains information about the host.
        
        protocol : ``AbstractProtocolBase``
            The protocol to add.
        
        keep_alive_timeout : `float`
            How long the connection can be reused.
        
        performed_requests : `int`
            The amount of performed requests on the connection.
        """
        protocols_by_host = self.protocols_by_host
        try:
            protocol_basket = protocols_by_host[key]
        except KeyError:
            protocol_basket = ProtocolBasket(key)
            protocols_by_host[key] = protocol_basket
        
        protocol_basket.add_available_protocol(protocol, LOOP_TIME(), keep_alive_timeout, performed_requests)
    
    
    def release_used_protocol(self, key, protocol):
        """
        Removes the given used protocol from the connector.
        
        Parameters
        ----------
        key : ``ConnectionKey``
            A key which contains information about the host.
        
        protocol : ``AbstractProtocolBase``
            The connected protocol to the respective host.
        """
        if self.closed:
            return
        
        protocols_by_host = self.protocols_by_host
        try:
            protocol_basket = protocols_by_host[key]
        except KeyError:
            return
        
        protocol_basket.remove_used_protocol(protocol)
        if not protocol_basket:
            del protocols_by_host[key]
    
    
    def release(self, key, protocol, should_close, keep_alive_timeout, performed_requests):
        """
        Releases the given protocol from the connector.
        If the connection should not be closed, not closes it, instead stores it for future reuse.
        
        Parameters
        ----------
        key : ``ConnectionKey``
            A key which contains information about the host.
            
        protocol : ``AbstractProtocolBase``
            Protocol of the released connection.
        
        should_close : `bool`
            Whether the respective connection should be closed.
        
        keep_alive_timeout : `float`
            How long the connection can be reused.
        
        performed_requests : `int`
            The amount of performed requests on the connection.
        """
        if self.closed:
            return
        
        self.release_used_protocol(key, protocol)
        
        if should_close or self.force_close or protocol.should_close():
            transport = protocol.get_transport()
            protocol.close()
            if key.secure and (transport is not None):
                transport.abort()
            return
        
        self.add_available_protocol(key, protocol, keep_alive_timeout, performed_requests)
        
        if self.clean_up_handle is None:
            self.clean_up_handle = self.loop.call_after_weak(keep_alive_timeout, self._clean_up)
    
    
    def get_closest_expiration(self):
        """
        Returns the closest expiration time or `-1.0`.
        
        Returns
        -------
        closest_expiration : `float`
        """
        closest_expiration = -1.0
        
        for protocol_basket in self.protocols_by_host.values():
            expiration = protocol_basket.get_closest_expiration()
            if expiration == -1.0:
                continue
            
            if closest_expiration == -1.0:
                closest_expiration = expiration
                continue
            
            if expiration < closest_expiration:
                closest_expiration = expiration
                continue
        
        return closest_expiration
    
    
    async def create_connection(self, request):
        """
        Creates a new connection for the given request.
        
        This method is a coroutine.
        
        Parameters
        ----------
        request : ``ClientRequest``
            Request to create connection for.
        
        Returns
        -------
        protocol : ``AbstractProtocolBase``
            The created protocol connected to the respective host.
        
        Raises
        ------
        NotImplementedError
            ``ConnectorBase`` do not implements ``.create_connection``. Subclasses should do it.
        """
        raise NotImplementedError



__all__ = ()

from re import compile as re_compile


KEEP_ALIVE_HEADER_RP = re_compile('\\s*([^ \\t=]+)\\s*=\\s*([^ \\t,$]+)\\s*(?:,|$)')

KEEP_ALIVE_CONNECTION_TIMEOUT_DEFAULT = 15.0
KEEP_ALIVE_MAX_REQUESTS_DEFAULT = 0

KEEP_ALIVE_CONNECTION_TIMEOUT_KEY = 'timeout'
KEEP_ALIVE_MAX_REQUESTS_KEY = 'max'



from .async_executing import *
from .event import *
from .future import *
from .future_chaining import *
from .future_states import *
from .future_wrapper_async import *
from .future_wrapper_base import *
from .future_wrapper_sync import *
from .handle_cancellers import *
from .locks import *
from .queues import *
from .task import *
from .task_group import *
from .task_suppression import *
from .task_thread_switcher import *
from .timeouting import *


__all__ = (
    *async_executing.__all__,
    *event.__all__,
    *future.__all__,
    *future_chaining.__all__,
    *future_states.__all__,
    *future_wrapper_async.__all__,
    *future_wrapper_base.__all__,
    *future_wrapper_sync.__all__,
    *handle_cancellers.__all__,
    *locks.__all__,
    *queues.__all__,
    *task.__all__,
    *task_group.__all__,
    *task_suppression.__all__,
    *task_thread_switcher.__all__,
    *timeouting.__all__,
)



__all__ = ('get_familiar_variable_names',)

from .helpers_familiarity import _get_familiar_names
from .helpers_variables import _get_variable_names


def get_familiar_variable_names(frame, name):
    """
    Gets all the familiar attribute names of the variable to the given one.
    
    Parameters
    ----------
    frame : `None | FrameProxyBase`
        Frame to get variable names from.
    name : `str`
        Name to check for.
    
    Returns
    -------
    variable_exists_just_was_not_set : `bool`
    familiar_variable_names : `None | list<str>`
    """
    variable_names = _get_variable_names(frame)
    
    try:
        variable_names.remove(name)
    except KeyError:
        variable_exists_just_was_not_set = False
    else:
        variable_exists_just_was_not_set = True
    
    familiar_variable_names = _get_familiar_names(variable_names, name)
    if not familiar_variable_names:
        familiar_variable_names = None
    
    return variable_exists_just_was_not_set, familiar_variable_names



__all__ = ('LineCacheSession',)

from .cache_constants import LINE_CACHE_SESSIONS


class LineCacheSession:
    """
    Keeps file caches alive till the length of a session.
    
    Attributes
    ----------
    memorized_file_infos : `set<FIleInfo>`
        The memorised file infos.
    """
    __slots__ = ('memorized_file_infos',)
    
    def __new__(cls):
        """
        Creates a new line cache context.
        """
        self = object.__new__(cls)
        self.memorized_file_infos = set()
        return self
    
    
    def __repr__(self):
        """Returns the line cache context's representation."""
        repr_parts = ['<', type(self).__name__]
        
        repr_parts.append(' memorized_file_infos = ')
        repr_parts.append(repr(self.memorized_file_infos))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __enter__(self):
        """
        Enters the context manager adding self to the active contexts.
        """
        LINE_CACHE_SESSIONS.add(self)
        return self
    
    
    def __exit__(self, exception_type, exception_value, exception_traceback):
        """
        Exists the context manager removing self from the active contexts.
        """
        LINE_CACHE_SESSIONS.remove(self)
        return False



__all__ = ('Proxy',)

from ssl import SSLContext

from ..utils import IgnoreCaseMultiValueDictionary, RichAttributeErrorBaseType
from ..web_common import BasicAuthorization, URL, freeze_headers

from .ssl_fingerprint import SSLFingerprint


class Proxy(RichAttributeErrorBaseType):
    """
    Represents a proxy to do request through.
    
    Attributes
    ----------
    _hash_cache : `None | int`
        Cache field for `hash(self)`.
    
    authorization : `None | BasicAuthorization`
        Authorization header to send.
    
    headers : `None | IgnoreCaseMultiValueDictionary`
        Additional to send to the proxy.
    
    ssl_context : `None | SSLContext``
        The proxy connection's ssl type.
    
    ssl_fingerprint : `None | SSLFingerprint`
        Alternative way to accept ssl or to block it depending whether the fingerprint is the same or changed.
    
    url : ``URL``
        The url to proxy through.
    """
    __slots__ = ('_hash_cache', 'authorization', 'headers', 'ssl_context', 'ssl_fingerprint', 'url')
    
    def __new__(cls, url, *, authorization = ..., headers = ..., ssl_context = ..., ssl_fingerprint = ...):
        # url
        url = URL(url)
        if url.host is None:
            raise ValueError(f'Host could not be detected from url: {url!r}')
        
        # authorization
        if authorization is ...:
            authorization = None
        elif (authorization is not None) and (not isinstance(authorization, BasicAuthorization)):
            raise TypeError(
                f'`authorization` can be be given as `None`, {BasicAuthorization.__name__}. '
                f'Got {type(authorization).__name__}; {authorization!r}.'
            )
        
        # headers
        if headers is ...:
            headers = None
        elif (headers is None):
            headers = None
        else:
            headers = IgnoreCaseMultiValueDictionary(headers)
            if not headers:
                headers = None
        
        # ssl_context
        if ssl_context is ...:
            ssl_context = None
        elif (ssl_context is not None) and (not isinstance(ssl_context, SSLContext)):
            raise TypeError(
                f'`ssl_context` can be given as `None`, `{type(SSLContext).__name__}`'
            )
    
        # ssl_fingerprint
        if ssl_fingerprint is ...:
            ssl_fingerprint = None
        elif (ssl_fingerprint is not None) and (not isinstance(ssl_fingerprint, SSLFingerprint)):
            raise TypeError(
                f'`ssl_fingerprint` can be given as `None`, `{type(SSLFingerprint).__name__}`'
            )
        
        # Construct
        self = object.__new__(cls)
        self._hash_cache = None
        self.authorization = authorization
        self.headers = headers
        self.ssl_context = ssl_context
        self.ssl_fingerprint = ssl_fingerprint
        self.url = url
        return self
    
    
    def __repr__(self):
        """Returns the proxy's representation."""
        repr_parts = ['<', type(self).__name__]
        
        # url
        repr_parts.append(' url = ')
        repr_parts.append(repr(self.url))
        
        # authorization
        authorization = self.authorization
        if (authorization is not None):
            repr_parts.append(', authorization = ')
            repr_parts.append(repr(authorization))
        
        # headers
        headers = self.headers
        if (headers is not None):
            repr_parts.append(', headers = ')
            repr_parts.append(repr(headers))
        
        # ssl_context
        ssl_context = self.ssl_context
        if (ssl_context is not None):
            repr_parts.append(', ssl_context = ')
            repr_parts.append(repr(ssl_context))
        
        # ssl_fingerprint
        ssl_fingerprint = self.ssl_fingerprint
        if (ssl_fingerprint is not None):
            repr_parts.append(', ssl_fingerprint = ')
            repr_parts.append(repr(ssl_fingerprint))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the proxy's hash value."""
        hash_value = self._hash_cache
        if (hash_value is None):
            hash_value = self._create_hash()
            self._hash_cache = hash_value
        
        return hash_value
    
    
    def _create_hash(self):
        """
        Generates the proxy's hash.
        
        Returns
        -------
        hash_value : `int`
        """
        hash_value = 0
        
        # authorization
        authorization = self.authorization
        if (authorization is not None):
            hash_value ^= hash(authorization)
        
        # headers
        headers = freeze_headers(self.headers)
        if (headers is not None):
            hash_value ^= hash(headers)
        
        # ssl_context
        ssl_context = self.ssl_context
        if (ssl_context is not None):
            hash_value ^= hash(ssl_context)
        
        # ssl_fingerprint
        ssl_fingerprint = self.ssl_fingerprint
        if (ssl_fingerprint is not None):
            hash_value ^= hash(ssl_fingerprint)
       
        # url
        hash_value ^= hash(self.url)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two proxies are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # authorization
        if self.authorization != other.authorization:
            return False
        
        # headers
        if self.headers != other.headers:
            return False
        
        # ssl_context
        if self.ssl_context != other.ssl_context:
            return False
        
        # ssl_fingerprint
        if self.ssl_fingerprint != other.ssl_fingerprint:
            return False
        
        # url
        if self.url != other.url:
            return False
        
        return True
    
    
    @property
    def host(self):
        """
        Returns the request's host.
        
        Returns
        -------
        host : `str`
        """
        return self.url.host
    
    
    @property
    def port(self):
        """
        Returns the request's port.
        
        Returns
        -------
        port : `None | int`
        """
        return self.url.port
    
    
    def is_secure(self):
        """
        Returns whether the request is secure.
        
        Returns
        -------
        is_secure : `bool`
        """
        scheme = self.url.scheme
        if scheme in ('http', 'ws'):
            return False
        
        if scheme in ('https', 'wss'):
            return True
        
        # If we have any other scheme let the user decide whether it is ssl or nah.
        return (self.ssl_context is not None)



__all__ = ('get_familiar_instance_attribute_names',)

from .helpers_familiarity import _get_familiar_names


def _get_instance_attributes(instance):
    """
    Gets the attribute names of the given instance.
    
    Parameter
    ---------
    instance : `object`
        The object to get its attribute names of.
    
    Returns
    -------
    attribute_names : `set<str>`
    """
    attribute_names = {*object.__dir__(instance)}
    
    attribute_names_getter = type(instance).__dir__
    if (attribute_names_getter is not None) and (attribute_names_getter is not object.__dir__):
        try:
            custom_names = attribute_names_getter(instance)
        except Exception:
            pass
        else:
            attribute_names.update(custom_names)
    
    attribute_names.update()
    return attribute_names


def get_familiar_instance_attribute_names(instance, name):
    """
    Gets all the familiar attribute names of the instance to the given one.
    
    Parameters
    ----------
    instance : `object`
        The object to check.
    name : `str`
        The attribute's name to check.
    
    Returns
    -------
    attribute_exists_just_was_not_set : `bool`
    familiar_attribute_names : `list<str>`
    """
    names = _get_instance_attributes(instance)
    
    try:
        names.remove(name)
    except KeyError:
        attribute_exists_just_was_not_set = False
    else:
        attribute_exists_just_was_not_set = True
    
    familiar_attribute_names = _get_familiar_names(names, name)
    if not familiar_attribute_names:
        familiar_attribute_names = None
    
    return attribute_exists_just_was_not_set, familiar_attribute_names



__all__ = ('FrameProxyVirtual',)

from ...docs import copy_docs

from .frame_proxy_base import FrameProxyBase


class FrameProxyVirtual(FrameProxyBase):
    """
    Virtual frame proxy.
    
    Attributes
    ----------
    alike_count : `int`
        How much alike frames are in the same exception.
    builtins : `None | dict<str, object>`
        The frame's builtins. Can be `None` if the variables were already omitted.
    code : `None | CodeType`
        The frame's code.
    expression_info : `None | ExpressionInfo`
        Additional expression information.
    file_name : `str`
        The represented frame's file name. Can be empty string.
    globals : `None | dict<str, object>`
        The frame's builtins. Can be `None` if the variables were already omitted.
    line_index : `int`
        The frame's current line index in Python source code.
    instruction_index : `None`
        The frame's last attempted instruction index in the bytecode.
    locals : `None | dict<str, object>`
        The local variables of the frame. Can be `None` if the variables were already omitted.
    name : `str`
        The frame's function's name. Can be empty string.
    tracing_function : `None | FunctionType`
        Tracing function for the traceback's frame. Defaults to `None`.
    """
    __slots__ = (
        'builtins', 'code', 'file_name', 'globals', 'instruction_index', 'line_index', 'locals', 'name',
        'tracing_function'
    )
    
    def __new__(cls, frame_proxy, *, with_variables = False):
        """
        Creates a new frame proxy.
        
        Parameters
        ----------
        frame_proxy : ``FrameProxyBase``
            The frame to freeze.
        with_variables : `bool` = `False`, Optional (Keyword only)
            Whether variables should also be inherited.
        """
        if isinstance(frame_proxy, cls):
            if frame_proxy.has_variables() <= with_variables:
                return frame_proxy
        
        self = object.__new__(cls)
        self.alike_count = 0
        self.builtins = frame_proxy.builtins if with_variables else None
        self.code = frame_proxy.code
        self.expression_info = frame_proxy.expression_info
        self.file_name = frame_proxy.file_name
        self.globals = frame_proxy.globals if with_variables else None
        self.instruction_index = frame_proxy.instruction_index
        self.line_index = frame_proxy.line_index
        self.locals = frame_proxy.locals if with_variables else None
        self.name = frame_proxy.name
        self.tracing_function = frame_proxy.tracing_function if with_variables else None
        return self
    
    
    @classmethod
    def from_fields(
        cls,
        *,
        builtins = ...,
        code = ...,
        expression_info = ...,
        file_name = ...,
        globals = ...,
        instruction_index = ...,
        line_index = ...,
        locals = ...,
        name = ...,
        tracing_function = ...,
    ):
        """
        Creates a new frame proxy from the given values.
        
        Parameters
        ----------
        builtins : `None | dict<str, object>`, Optional (Keyword only)
            The frame's builtins.
        code : `None | CodeType`, Optional (Keyword only)
            The frame's code.
        expression_info : `None | ExpressionInfo`, Optional (Keyword only)
            Additional expression information.
        file_name : `str`, Optional (Keyword only)
            The represented file's name.
        globals : `None | dict<str, object>`, Optional (Keyword only)
            The frame's builtins.
        instruction_index : `None`, Optional (Keyword only)
            The frame's last attempted instruction index in the bytecode.
        line_index : `int`, Optional (Keyword only)
            The frame's current line index in Python source code.
        locals : `None | dict<str, object>`, Optional (Keyword only)
            The local variables of the frame.
        name : `str`
            The frame's function's name.
        tracing_function : `None | FunctionType`, Optional (Keyword only)
            Tracing function for the traceback's frame.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.alike_count = 0
        self.builtins = None if builtins is ... else builtins
        self.code = None if code is ... else code
        self.expression_info = None if expression_info is ... else expression_info
        self.file_name = '' if file_name is ... else file_name
        self.globals = None if globals is ... else globals
        self.instruction_index = 0 if instruction_index is ... else instruction_index
        self.line_index = 0 if line_index is ... else line_index
        self.locals = None if locals is ... else locals
        self.name = '' if name is ... else name
        self.tracing_function = None if tracing_function is ... else tracing_function
        return self
    
    
    @copy_docs(FrameProxyBase.has_variables)
    def has_variables(self):
        if self.builtins is not None:
            return True
        
        if self.globals is not None:
            return True
        
        if self.locals is not None:
            return True
        
        return False



__all__ = ()

from warnings import warn

from ..core.top_level import write_exception_async
from ..utils import RichAttributeErrorBaseType
from ..web_common.constants import KEEP_ALIVE_MAX_REQUESTS_DEFAULT


class Connection(RichAttributeErrorBaseType):
    """
    Reusable connection used by connectors to remember their details.
    
    Attributes
    ----------
    callbacks : `list<callable>`
        Callable-s to run when the connection is ``.close``-d, ``.release``-d or ``.detach``-ed.
        They should accept no parameters.
    
    connector : ``ConnectorBase``
        The respective connector of the connection.
    
    key : ``ConnectionKey``
        A key which contains information about the host.
    
    performed_requests : `int`
        The amount of already performed requests.
    
    protocol : `None | HttpReadWriteProtocol`
        The connection's actual protocol.
        Set as `None` when the connection is ``.close``-d, ``.release``-d or ``.detach``-ed.
    """
    __slots__ = ('callbacks', 'connector', 'key', 'performed_requests', 'protocol',)
    
    def __new__(cls, connector, key, protocol, performed_requests):
        """
        Creates a new connection with the given parameters.
        
        Parameters
        ----------
        connector : ``ConnectorBase``
            The respective connector of the connection.
        
        key : ``ConnectionKey``
            A key to identify the connection.
        
        protocol : ``HttpReadWriteProtocol``
            The connection's actual protocol.
        
        performed_requests : `int`
            The amount of already performed requests.
        """
        self = object.__new__(cls)
        self.callbacks = []
        self.connector = connector
        self.key = key
        self.protocol = protocol
        self.performed_requests = performed_requests
        return self
    
    
    def __repr__(self):
        """Returns the representation of the connection."""
        repr_parts = ['<', type(self).__name__]
        
        # key
        repr_parts.append(' to ')
        repr_parts.append(repr(self.key))
        
        # closed
        if self.is_closed():
            repr_parts.append(' closed')
        
        # performed_requests
        performed_requests = self.performed_requests
        if performed_requests:
            repr_parts.append(', performed_requests = ')
            repr_parts.append(repr(performed_requests))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __del__(self):
        """Releases the connector if not yet released."""
        connector = self.connector
        if connector.loop.running:
            protocol = self.protocol
            if (protocol is not None):
                connector.release(self.key, protocol, True, -1.0, 0)
    
    
    def get_transport(self):
        """
        Returns the connection's ``.protocol``'s transport if applicable.
        
        Returns
        -------
        transport : `None | AbstractTransportLayerBase`
            Defaults to `None`.
        """
        protocol = self.protocol
        if (protocol is not None):    
            return protocol.get_transport()
    
    
    @property
    def transport(self):
        """
        Deprecated and will be removed in 2025 April. Please use ``.get_transport`` instead.
        """
        warn(
            (
                f'`{type(self).__name__}.transport` is deprecated and will be removed in 2025 April. '
                f'Please use `.get_transport()` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        return self.get_transport()
    
    
    def add_callback(self, callback):
        """
        Adds a callback to the connection. If the connection is already closed runs it instantly.
        
        Parameters
        ----------
        callback : `callable`
            Callable to run when the connection is ``.close``-d, ``.release``-d or ``.detach``-ed.
            Should accept no parameters.
        """
        if not self.is_closed():
            self.callbacks.append(callback)
            return
        
        try:
            callback()
        except BaseException as err:
            write_exception_async(
                err,
                [
                    'Exception occurred at ',
                    repr(self),
                    '.add_callback\nAt running ',
                    repr(callback),
                    '\n',
                ],
                loop = self.connector.loop,
            )
    
    
    def _run_callbacks(self):
        """
        Runs the added callbacks of the connection.
        """
        callbacks = self.callbacks
        while callbacks:
            callback = callbacks.pop()
            
            try:
                callback()
            except BaseException as err:
                write_exception_async(
                    err,
                    [
                        'Exception occurred at ',
                        repr(self),
                        '._run_callbacks\nAt running ',
                        repr(callback),
                        '\n',
                    ],
                    loop = self.connector.loop,
                )
    
    
    def close(self):
        """
        Closes the connection by running it's callbacks and releasing with the intent of closing.
        """
        self._run_callbacks()
        
        protocol = self.protocol
        if (protocol is not None):
            self.protocol = None
            self.connector.release(self.key, protocol, True, -1.0, 0)
    
    
    def release(self, keep_alive_info):
        """
        Closes the connection by running it's callbacks and releasing it.
        """
        self._run_callbacks()
        
        protocol = self.protocol
        if (protocol is None):
            return
        
        self.protocol = None
        performed_requests = self.performed_requests + 1
        
        while True:
            should_close = protocol.should_close()
            if should_close:
                connection_keep_alive_timeout = -1.0
                break
            
            # no keep alive -> close
            if keep_alive_info is None:
                should_close = True
                connection_keep_alive_timeout = -1.0
                break
            
            # If we maxed our keep alive -> close
            max_requests = keep_alive_info.max_requests
            if (max_requests != KEEP_ALIVE_MAX_REQUESTS_DEFAULT) and (performed_requests >= max_requests):
                should_close = True
                connection_keep_alive_timeout = -1.0
                break
            
            should_close = False
            connection_keep_alive_timeout = keep_alive_info.connection_timeout
            break
        
        self.connector.release(self.key, protocol, should_close, connection_keep_alive_timeout, performed_requests)
    
    
    def detach(self):
        """
        Detaches the connection from it ``.connector``.
        
        Note, that with detaching, the connection set it's ``.protocol`` as `None`, so you should save an instance of
        it down before.
        """
        self._run_callbacks()
        
        protocol = self.protocol
        if (protocol is not None):
            self.protocol = None
            self.connector.release_used_protocol(self.key, protocol)
   
   
    def is_closed(self):
        """
        Returns whether the connection is closed.
        
        Returns
        -------
        closed : `bool`
        """
        protocol = self.protocol
        if protocol is None:
            return True
        
        if protocol.get_transport() is None:
            return True
        
        return False
    
    
    @property
    def closed(self):
        """
        Deprecated and will be removed in 2025 April. Please use ``.is_closed`` instead.
        """
        warn(
            (
                f'`{type(self).__name__}.closed` is deprecated and will be removed in 2025 April. '
                f'Please use `.is_closed()` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        return self.is_closed()



from .familiar_instance_attribute_names import *
from .familiar_variable_names import *
from .helpers_familiarity import *
from .helpers_variables import *
from .matching_variable_name import *
from .matching_variable_names_with_attribute import *


__all__ = (
    *familiar_instance_attribute_names.__all__,
    *familiar_variable_names.__all__,
    *helpers_familiarity.__all__,
    *helpers_variables.__all__,
    *matching_variable_name.__all__,
    *matching_variable_names_with_attribute.__all__,
)



__all__ = ('add_console_input',)

from .cache_constants import (
    CONSOLE_INPUT_CACHE, CONSOLE_INPUT_MAX_SIZE, CONSOLE_INPUT_CACHE_OUTFLOW_CHECK, FILE_INFO_CACHE
)
from .file_info import FileInfo


def add_console_input(file_name, content):
    """
    Adds a console input. Ignores empty input.
    
    Parameters
    ----------
    file_name : `None | str`
        File name to add the input as.
    content : `None | str`
        Input content to add.
    """
    if (file_name is None) or (not file_name) or (content is None) or (not content):
        return
    
    size = len(content)
    lines = content.splitlines()
    
    file_info = FileInfo(file_name, size, 0.0, lines, False, True)
    
    # Put file to permanent cache
    CONSOLE_INPUT_CACHE[file_name] = file_info
    FILE_INFO_CACHE[file_name] = file_info
    
    if len(CONSOLE_INPUT_CACHE) % CONSOLE_INPUT_CACHE_OUTFLOW_CHECK == 0:
        check_console_cache_overflow()


def check_console_cache_overflow():
    """
    Checks whether the console cache is overflown. If yes clears the overflow.
    """
    total = 0
    to_delete = None
    
    for file_info in reversed(CONSOLE_INPUT_CACHE.values()):
        total += file_info.file_size
        if total < CONSOLE_INPUT_MAX_SIZE:
            continue
        
        if to_delete is None:
            to_delete = []
        
        to_delete.append(file_info.file_name)
    
    
    if (to_delete is not None):
        for file_name in to_delete:
            del CONSOLE_INPUT_CACHE[file_name]



__all__ = ('PayloadStream', )

from collections import deque as Deque

from ...utils import RichAttributeErrorBaseType, include

from ..traps import Future


write_exception_async = include('write_exception_async')


STREAM_FLAG_WAIT_WHOLE = 1 << 0
STREAM_FLAG_WAIT_CHUNK = 1 << 1
STREAM_FLAG_DONE_SUCCESS = 1 << 2
STREAM_FLAG_DONE_EXCEPTION = 1 << 3
STREAM_FLAG_DONE_CANCELLED = 1 << 4
STREAM_FLAG_DONE_ABORTED = 1 << 5

STREAM_FLAG_WAIT_ANY = STREAM_FLAG_WAIT_WHOLE | STREAM_FLAG_WAIT_CHUNK
STREAM_FLAG_DONE_RAISE_ANY = STREAM_FLAG_DONE_EXCEPTION | STREAM_FLAG_DONE_CANCELLED | STREAM_FLAG_DONE_ABORTED
STREAM_FLAG_DONE_ANY = STREAM_FLAG_DONE_SUCCESS | STREAM_FLAG_DONE_RAISE_ANY

    
def _get_payload_stream_flags_name(flags):
    """
    Get flags name.
    
    Parameters
    ----------
    flags : `int`
        Flags to get their name of.
    
    Returns
    -------
    name : `str`
    """
    if not flags:
        return 'none'
    
    output_parts = []
    
    if flags & STREAM_FLAG_WAIT_WHOLE:
        output_parts.append('wait~whole')
    
    if flags & STREAM_FLAG_WAIT_CHUNK:
        output_parts.append('wait~chunk')
    
    if flags & STREAM_FLAG_DONE_SUCCESS:
        output_parts.append('done~success')
    
    if flags & STREAM_FLAG_DONE_EXCEPTION:
        output_parts.append('done~exception')
    
    if flags & STREAM_FLAG_DONE_CANCELLED:
        output_parts.append('done~cancelled')
    
    if flags & STREAM_FLAG_DONE_ABORTED:
        output_parts.append('done~aborted')
    
    return ', '.join(output_parts)


def _create_payload_stream_done_exception(flags):
    """
    Creates an exception used at payload reader cancellation.
    
    Parameters
    ----------
    flags : `int`
        Payload reader flags.
    
    Returns
    -------
    exception : ``ConnectionError``
    """
    if flags & STREAM_FLAG_DONE_CANCELLED:
        message = 'Payload reader cancelled.'
    else:
        message = 'Payload reader aborted.'
    
    return ConnectionError(message)


class PayloadStream(RichAttributeErrorBaseType):
    """
    Payload stream.
    
    Attributes
    ----------
    _chunks : `Deque<bytes | memoryview>`
        Cached chunks.
    
    _done_callbacks : `None | list<callable>`
        Functions to call when the stream ended.
    
    _flags : `int`
        Bitwise flags storing the state of the stream.
    
    _exception : `None | BaseException`
        Payload reader exception.
    
    _protocol : ``ReadProtocolBase``
        The parent protocol.
    
    _waiter : `None | Future`
        Payload waiter used when waiting for a new chunk or for all chunks respectively.
    """
    __slots__ = ('_chunks', '_done_callbacks', '_exception', '_flags', '_protocol', '_waiter')
    
    def __new__(cls, protocol):
        """
        Creates a new payload stream.
        
        Parameters
        ----------
        protocol : ``AbstractProtocolBase``
            The parent protocol.
        """
        self = object.__new__(cls)
        self._chunks = Deque()
        self._done_callbacks = None
        self._exception = None
        self._flags = 0
        self._protocol = protocol
        self._waiter = None
        return self
    
    
    def __repr__(self):
        """Returns the payload streams representation."""
        repr_parts = ['<', type(self).__name__]
        
        repr_parts.append(' total_size = ')
        repr_parts.append(repr(self.get_total_size()))
        
        repr_parts.append(', flags = ')
        repr_parts.append(_get_payload_stream_flags_name(self._flags))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def get_total_size(self):
        """
        Gets the total size of the payload stream.
        
        Returns
        -------
        size : `int`
        """
        size = 0
        for chunk in self._chunks:
            size += len(chunk)
        
        return size
    
    
    def get_buffer_size(self):
        """
        Gets the buffer size of the payload stream. If we are expecting to receive the whole payload it returns `0`.
        
        Returns
        -------
        size : `int`
        """
        if self._flags & STREAM_FLAG_WAIT_WHOLE:
            return 0
        
        return self.get_total_size()
    
    
    def set_done_success(self):
        """
        Sets the payload stream to be done with success.
        
        Returns
        -------
        success : `bool`
        """
        flags = self._flags
        if flags & STREAM_FLAG_DONE_ANY:
            return False
        
        self._flags |= STREAM_FLAG_DONE_SUCCESS
        
        waiter = self._waiter
        if (waiter is not None):
            self._waiter = None
            waiter.set_result_if_pending(None)
        
        self._run_done_callbacks()
        return True
    
    
    def set_done_cancelled(self):
        """
        Sets the payload stream to be done by cancellation.
        
        Returns
        -------
        success : `bool`
        """
        flags = self._flags
        if flags & STREAM_FLAG_DONE_ANY:
            return False
        
        flags |= STREAM_FLAG_DONE_CANCELLED
        self._flags = flags
        
        waiter = self._waiter
        if (waiter is not None):
            self._waiter = None
            waiter.set_exception_if_pending(_create_payload_stream_done_exception(flags))
        
        self._chunks.clear()
        self._run_done_callbacks()
        return True
    
    
    def set_done_exception(self, exception):
        """
        Sets the payload stream to be done with exception.
        
        Parameters
        ----------
        exception : `BaseException`
            Exception to set.
        
        Returns
        -------
        success : `bool`
        """
        flags = self._flags
        if flags & STREAM_FLAG_DONE_ANY:
            return False
        
        self._flags |= STREAM_FLAG_DONE_EXCEPTION
        
        waiter = self._waiter
        if (waiter is not None):
            self._waiter = None
            waiter.set_exception_if_pending(exception)
        
        self._exception = exception
        self._chunks.clear()
        self._run_done_callbacks()
        return True
    
    
    def add_received_chunk(self, chunk):
        """
        Adds a received chunk to the payload stream's queue.
        
        Parameters
        ----------
        chunk : `bytes | memoryview`
            Received chunk.
        
        Returns
        -------
        success : `bool`
        """
        flags = self._flags
        if flags & STREAM_FLAG_DONE_ANY:
            return False
        
        if flags & STREAM_FLAG_WAIT_CHUNK:
            waiter = self._waiter
            if (waiter is not None):
                self._waiter = None
                if waiter.set_result_if_pending(chunk):
                    return True
        
        self._chunks.append(chunk)
        return True
    
    
    def add_done_callback(self, callback):
        """
        Adds a done callback to the payload stream.
        
        Parameters
        ----------
        callback : `callable`
            Callback to run when the stream is done.
        """
        if self._flags & STREAM_FLAG_DONE_ANY:
            self._run_done_callback(callback)
        else:
            done_callbacks = self._done_callbacks
            if (done_callbacks is None):
                done_callbacks = []
                self._done_callbacks = done_callbacks
            done_callbacks.append(callback)
    
    
    def _run_done_callbacks(self):
        """
        Runs the added done callbacks of the payload stream.
        """
        done_callbacks = self._done_callbacks
        if done_callbacks is None:
            return
        
        while done_callbacks:
            self._run_done_callback(done_callbacks.pop())
        
        self._done_callbacks = None
    
    
    def _run_done_callback(self, callback):
        """
        Runs a single callback.
        
        Parameters
        ----------
        callback : `callable`
            The callback to run.
        """
        try:
            callback(self)
        except BaseException as err:
            write_exception_async(
                err,
                [
                    'Exception occurred at ',
                    repr(self),
                    '._run_done_callbacks\nAt running ',
                    repr(callback),
                    '\n',
                ],
                loop = self._protocol._loop,
            )
    
    
    def _abort(self):
        """
        Aborts the payload stream.
        
        Returns
        -------
        success : `bool`
        """
        flags = self._flags
        if flags & STREAM_FLAG_DONE_ANY:
            return False
        
        flags |= STREAM_FLAG_DONE_ABORTED
        self._flags = flags
        
        waiter = self._waiter
        if (waiter is not None):
            self._waiter = None
            if waiter.is_pending():
                waiter.set_exception(_create_payload_stream_done_exception(flags))
        
        self._run_done_callbacks()
        return True
    
    
    def is_aborted(self):
        """
        Returns whether the payload stream was aborted.
        
        Returns
        -------
        aborted : `bool`
        """
        return True if self._flags & STREAM_FLAG_DONE_ABORTED else False
    
    
    def _check_raise_flags(self):
        """
        Checks the raise flags of the payload reader. Raises if applicable.
        
        Raises
        ------
        ConnectionError
            Payload reading cancelled.
        BaseException
            Payload reader exception.
        """
        flags = self._flags
        if flags & STREAM_FLAG_DONE_RAISE_ANY:
            exception = self._exception
            if (exception is None):
                exception = _create_payload_stream_done_exception(flags)
            raise exception
    
    
    def _check_wait_flags(self, expected_flag):
        """
        Checks the wait flags of the payload reader.
        
        Raises
        ------
        RuntimeError
            Payload reader would go to an invalid state.
        """
        flags = self._flags
        wait_flags = flags & STREAM_FLAG_WAIT_ANY
        if wait_flags == expected_flag:
            pass
        
        elif not wait_flags:
            self._flags = flags | expected_flag
        
        else:
            raise RuntimeError(
                f'Payload stream is in invalid state to: {_get_payload_stream_flags_name(expected_flag)!s}. '
                f'Current state: {_get_payload_stream_flags_name(flags)!s}.'
            )
    
    
    def __await__(self):
        """
        Awaits the payload reader.
        
        This method is an awaitable generator.
        
        Returns
        -------
        data : `bytes`
        
        Raises
        ------
        ConnectionError
            Payload reading cancelled.
        RuntimeError
            Payload reader would go to an invalid state.
        BaseException
            Payload reader exception.
        """
        self._check_wait_flags(STREAM_FLAG_WAIT_WHOLE)
        self._check_raise_flags()
        
        if not self._flags & STREAM_FLAG_DONE_SUCCESS:
            waiter = self._waiter
            if (waiter is None):
                self._protocol._resume_reading()
                waiter = Future(self._protocol._loop)
                self._waiter = waiter
            
            try:
                yield from waiter
            except:
                self._abort()
                raise
        
        chunks = self._chunks
        data = b''.join(chunks)
        chunks.clear()
        return data
    
    
    async def __aiter__(self):
        """
        Iterates over the received chunks of the payload reader as they come.
        
        This method is an coroutine generator.
        
        Yields
        -------
        chunk : `bytes | memoryview`
        
        Raises
        ------
        ConnectionError
            Payload reading cancelled.
        RuntimeError
            Payload reader would go to an invalid state.
        BaseException
            Payload reader exception.
        """
        chunks = self._chunks
        self._check_wait_flags(STREAM_FLAG_WAIT_CHUNK)
        
        while True:
            self._check_raise_flags()
            
            if chunks:
                chunk = chunks.popleft()
                self._protocol._resume_reading()
            
            else:
                self._check_raise_flags()
                if self._flags & STREAM_FLAG_DONE_SUCCESS:
                    return
                
                waiter = self._waiter
                if (waiter is None):
                    waiter = Future(self._protocol._loop)
                    self._waiter = waiter
                
                try:
                    chunk = await waiter
                except:
                    self._abort()
                    raise
                
                if chunk is None:
                    return
            
            try:
                yield chunk
            except:
                self._abort()
                raise
            
            continue
